#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Checking Job Creation
#for the record, Emos alternative syntax is,

#!/usr/bin/env python2.7
import os
from ecf import *
print "suite creation"
defs = Defs()
defs.suite("test").add(
    Variables(ECF_HOME= os.path.join(os.getenv("HOME"), "course")),
    Task("t1"))
print defs
print "Checking job creation: .ecf -> .job0"  
print defs.check_job_creation()
>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# Load the file
#Using "if __name__ == '__main__':" the script can be used both as a module (providing a function) or as a standalone script:

#!/usr/bin/env python2.7
import os
from ecf import *

def generate():
    # print "Creating suite definition"  
    return Defs().add(
        Suite("test").add(
            Variables(ECF_HOME= os.path.join(os.getenv("HOME"),  "course")),
            Task("t1")))
 
if __name__ == '__main__':
    defs = generate()      
    print defs
    print "Checking job creation: .ecf -> .job0"  
    print defs.check_job_creation()
    print "Saving definition to file 'test.def'"
    defs.save_as_defs("test.def")
    try:
        print "Load the in memory definition(defs) into the server"
        Client("localhost@%s" % os.getenv("ECF_PORT")).load(defs)
    except RuntimeError as e: print "Failed:", e
        
>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# families
#!/usr/bin/env python2.7
import os
from ecf import *
print "Creating suite definition"  
defs = Defs().add(Suite("test").add(
    Variable("ECF_HOME", os.path.join(os.getenv("HOME"),  "course")),
    Family("f1").add(
      Task("t1"),
      Task("t2"))))
print defs
print "Checking job creation: .ecf -> .job0"  
print defs.check_job_creation()
print "Saving definition to file 'test.def'"
defs.save_as_defs("test.def")
print "Client -> Server: delete, then load a new definition"  
try:
    ci = Client()
    ci.replace("/test", "test.def") # replace the suite into the server
    ci.begin_suite("test")    # start the suite
except RuntimeError as e:
    print "Failed:",   e

>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# ecflow variables
#!/usr/bin/env python2.7
import os
from ecf import *
home = os.path.join(os.getenv("HOME"), "course")
 
def create_family_f1():
    return Family("f1" ).add(
        Task("t1").add(Variables(SLEEP= 20)),
        Task("t2").add(Variables(SLEEP= 20)))
 
defs = Defs().add(Suite("test").add(
    Variables(ECF_INCLUDE= home,
              ECF_HOME=    home),
    create_family_f1() ))
 
print defs
defs.save_as_defs("test.def")

>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# variable inheritance
# tab constraint is relaxed with Emos ecf.py
#!/usr/bin/env python2.7
import os
from ecf import *
home = os.path.join(os.getenv("HOME"), "course")
def create_family_f1():
    return Family("f1" ).add(
        Variables(SLEEP= 20),
        Task("t1"),
        Task("t2"))
print "Creating suite definition"
defs = Defs().add(Suite("test").add(
    Variables(ECF_INCLUDE= home,
             ECF_HOME=    home),
     create_family_f1() ))
print defs

>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# triggers#!/usr/bin/env python2.7
import os
from ecf import *
home=  os.path.join(os.getenv("HOME"), "course")
def create_family_f1():
    return Family("f1" ).add(
        Variable(SLEEP= 20),
        Task("t1"),
        Task("t2").add(Trigger("t1 eq complete")))
print "Creating suite definition"
defs = Defs().add(Suite("test").add(
    Variables(ECF_INCLUDE= home,
              ECF_HOME=    home),
    create_family_f1() ))
print defs
defs.save_as_defs("test.def")

>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# add an event
#!/usr/bin/env python2.7
import os
from ecf import *
home = os.path.join(os.getenv("HOME"), "course")
def create_family_f1():
    return Family("f1").add(
        Variables(SLEEP= 20),
        Task("t1"),
        Task("t2").add(
            Trigger(["t1"]),
            Event("a"),
            Event("b")),
     
        Task("t3").add(Trigger("t2:a")),
        Task("t4").add(Trigger("t2:b")),
    )
print "Creating suite definition"
defs = Defs().add(
    Suite("test").add(
        Variables(ECF_INCLUDE= home,
                  ECF_HOME=    home),
        create_family_f1() ))
print defs
defs.save_as_defs("test.def")

>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# add a complete
#!/usr/bin/env python2.7
import os
from ecf import *
home = os.path.join(os.getenv("HOME"),  "course")
def create_family_f1():
    return Family("f1").add(
        Variable(SLEEP= 20),
        Task("t1"),
        Task("t2").add(Trigger("t1 eq complete"),
                       Event("a"),
                       Event("b")),
        Task("t3").add(Trigger("t2:a")),
        Task("t4").add(Trigger("t2 eq complete"), 
                       Complete("t2:b")  ))
       
print "Creating suite definition"  
defs = Defs().add(Suite("test").add(
    Variables(ECF_INCLUDE= home,
              ECF_HOME=    home),
    create_family_f1() ))
print defs
defs.save_as_defs("test.def")

>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# add a meter
#!/usr/bin/env python2.7
import os
from ecf import *
home= os.path.join(os.getenv("HOME"), "course")
def create_family_f1():
    return Family("f1").add(
        Variables(SLEEP= 20),
        Task("t1").add(Meter("progress", 1, 100, 90)),
     
        Task("t2").add(Trigger("t1 eq complete"),
                       Event("a"),
                       Event("b")),
     
        Task("t3").add(Trigger("t2:a")),
     
        Task("t4").add(Trigger("t2 eq complete"),
                       Complete("t2:b")),
      
        Task("t5").add(Trigger("t1:progress ge 30")),
        Task("t6").add(Trigger("t1:progress ge 60")),
        Task("t7").add(Trigger("t1:progress ge 90")))
    
print "Creating suite definition"  
defs = Defs().add(Suite("test").add(
    Variables(ECF_INCLUDE= home,
              ECF_HOME=    home),
    create_family_f1() ))
print defs
defs.save_as_defs("test.def")


>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# Time dependencies
# #!/usr/bin/env python2.7
import os

home = os.path.join(os.getenv("HOME"),  "course")
 
def create_family_f2():
    return Family("f2").add(
        Variables(SLEEP= 20),
        Task("t1").add(
          Time("00:30 23:30 00:30")), # start(hh:mm) end(hh:mm) increment(hh:mm)
        Task("t2").add(Day( "sunday" )),
        Task("t3").add(
            Date("1.*.*"),  # day month year, first of every month or every year
            Time("12:00")), # hour, minutes at 12 o'clock
    
        Task("t4").add(Time("+00:02")), # hour, minutes, relative to suite start
                                        # 2 minutes after family f2 start
        Task("t5").add(Time("00:02"))   # hour, minutes suite site
                                        # 2 minutes past midnight
)           
    
defs = Defs().add(Suite("test").add(
    Variables(ECF_INCLUDE= home,
              ECF_HOME=    home),
    create_family_f1(),
    create_family_f2(), ))
print defs
defs.save_as_defs("test.def")


>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# indentation
# Another approach is to use ecf.py module to benefit from object oriented polymorphism
#!/usr/bin/env python2.7
import os
import sys
sys.path.append('/home/ma/emos/def/o/def')
from ecf import *
print "Creating suite definition" 
defs = Defs().add(# Stream like definition
    Suite("test").add(
        Variables({ # a dictionary to detect duplicated variables
                "ECF_INCLUDE": os.getenv("HOME") + "/course",
                "ECF_HOME":    os.getenv("HOME") + "/course",}),
        Family("f1").add(
            Variable("SLEEP", "20"),
            Task("t1").add(Meter("progress", 1, 100, 90)),
            Task("t2").add(
                Trigger("t1 eq complete"),
                Event("a"),
                Event("b")),
            Task("t3").add(Trigger("t2:a")),
            Task("t4").add(Trigger("t2 eq complete"),
                           Complete("t2:b")),
            Task("t5").add(Trigger("t1:progress ge 30")),
            Task("t6").add(Trigger("t1:progress ge 60")),
            Task("t7").add(Trigger("t1:progress ge 90")),),
        Family("f2").add(
            Variable("SLEEP", "20"),
            Task("t1").add(Time( "00:30 23:30 00:30" )),
            Task("t2").add(Day( "sunday" )),
            Task("t3").add(Date("1.*.*"),
                           Time("12:00")),
            Task("t4").add(Time("+00:02")),
            Task("t5").add(Time("00:02")))))
             
out = file("test.def", "w")
print >>out, defs


>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
# label
#!/usr/bin/env python2.7
import os
from ecf import *
home = os.getenv("HOME") + "/course"
print "Creating suite definition"  
defs = Defs().add(Suite("test").add(
    Variables(ECF_INCLUDE= home,
              ECF_HOME=    home),
    Family("f3").add(
        Task("t1").add(
            Label("info","")))))
print defs
defs.save_as_defs("test.def")


>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
# repeat
#!/usr/bin/env python2.7
import os
from ecf import *
home = os.path.join(os.getenv("HOME"),  "course")
def create_family_f4():
    return Family("f4").add(
        Variables(SLEEP= 2),
        Repeat("NAME", ["a", "b", "c", "d", "e", "f" ], kind="enum"),
    
        Family("f5").add(
            Repeat( "VALUE", 1, 10, kind= "integer"),
            Task("t1").add(
                Repeat( "DATE", 20101230, 20110105, kind= "date"),
                Label("info", ""))))
     
print "Creating suite definition"  
defs = ecflow.Defs().add(Suite("test").add(
    Variables(ECF_INCLUDE= home,
              ECF_HOME=    home),
    create_family_f4(), ))
print defs
defs.save_as_defs("test.def")

>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# limits
#!/usr/bin/env python2.7
import os
from ecf import *
home= os.path.join(os.getenv("HOME"),  "course")
def create_family_f5() :
    return Family("f5").add(
        Inlimit("l1"),
        Variables(SLEEP= 20),
        [ Task( "t" + str(i))     for i in range(1, 10)] # list comprehension
    )
     
print "Creating suite definition"  
defs = Defs().add(Suite("test").add(
    Variables(ECF_INCLUDE= home,
              ECF_HOME=    home),
    Limit("l1", 2),
    create_family_f5() ))
print defs
defs.save_as_defs("test.def")

>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

#  Data acquiation solution
#!/usr/bin/env python2.7
import os
from ecf import * 
home = os.getenv("HOME") + "/course")  
defs = Defs().add(Suite("data_aquisition").add(
    Repeat(kind="day", step=1),
    Variables(ECF_HOME= home,
              ECF_INCLUDE= home,
              ECF_FILES= home + "/data",
              SLEEP= 2),
    [ Family(city).add(
        Task("archive").add(
            [ Family(obs_type).add(
                If( city in ("Exeter", "Toulouse", "Offenbach"),
                    Time("00:00 23:00 01:00")),
                If( city in ("Washington"),
                    Time("00:00 23:00 03:00")),
                If( city in ("Tokyo"),
                    Time("12:00")),
                If( city in ("Melbourne"),
                    Day( "monday" )),
                If( city in ("Montreal"),
                    Date("1.*.*")),
          
                Task("get"),
                Task("process").add(Trigger("get eq complete")),
                Task("store").add(Trigger("get eq complete")))
              for obs_type in ( "observations", "fields", "images" ) ] ))
        for city in ( "Exeter", "Toulouse", "Offenbach", "Washington", "Tokyo", "Melbourne", "Montreal" ) ]
    ))
print defs

>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# operational suite solution:

#!/usr/bin/env python2.7
import os
from ecf import *
home = os.getenv("HOME") + "/course"
cycle_triggers = None
last_step = { "12": 240,
              "00": 24, }
def cycle_trigger(cycle):
    if cycle == "12": return Trigger("./00==complete")
    return None
defs = Defs().add(Suite("operation_suite").add(
     Repeat(kind="day", step=1),
     Variables(ECF_HOME= home,
               ECF_INCLUDE= home,
               ECF_FILES= home + "/oper"),
    [ Family(str(cycle)).add(
        Variables(CYCLE=  cycle,
                  LAST_STEP= last_step[cycle]),
       
        cycle_trigger(cycle),
                  
        Family("analysis").add(
            Task("get_observations"),
            Task("run_analysis").add(Trigger(["get_observations", ])),
            Task("post_processing").add(Trigger(["run_analysis", ])),
        ),
  
        Family("forecast").add(
            Trigger("analysis == complete"),
            Task("get_input_data"),
            Task("run_forecast").add(
                Trigger(["get_input_data", ]),
                Meter("step", 0, last_step[cycle])),
        ),
        Family("archive").add(
            Family("analysis").add(
                Variables(TYPE= "analysis",
                          STEP= 0),
                Trigger(["../analysis/run_analysis", ]),
                Task("save"),
                [ Family("step_%02d" % i).add(
                    Variables(TYPE= "forecast",
                              STEP= i),
                    Trigger("../../forecast/run_forecast:step ge %d" % i),
                    Task("save"))
                  for i in range(6, last_step+1, 6) ]
            )
        )
    ) for cycle in ( 0 , 12 ) ] ))
print defs

!!!!!!! 
Baudouin Raoult
range(6, last_step, 6) will not return last_step.

>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

# back archiving solution
#!/usr/bin/env python2.7
import os
from ecf import *
home= os.getenv("HOME") + "/course"
defs = Defs().add(Suite("back_archiving").add(
    Repeat( kind= "day", step= 1),
    Variables(ECF_HOME= home,
              ECF_INCLUDE= home,
              ECF_FILES= home + "/back",
              SLEEP= 2),
    Limit("access", 2),
    [ Family(kind).add(
        Repeat( "DATE", 19900101, 19950712, kind="date"),
        Variables(KIND= kind),
        Task("get_old").add(Inlimit("access")),
        Task("convert").add(Trigger("get_old == complete")),
        Task("save_new").add(Trigger("convert == complete")))
      for kind in ( "analysis", "forecast", "climatology", "observations", "images" ) ]
))
print defs



import ecflow

try:
    ci = ecflow.Client()
    ci.sync_local()  # get the defs from the server, and place on ci
    defs = ci.get_defs()  # retrieve the defs from ci
    if len(defs) == 0:
        print("No suites in server, loading defs from disk")
        ci.load("test.def")

        print("Restarting the server. This starts job scheduling")
        ci.restart_server()
    else:
        print("read definition from disk and replace on the server")
        ci.replace("/test", "test.def")

    print("Begin the suite named 'test'")
    ci.begin_suite("test")

except RuntimeError as e:
    print("Failed:", e)



import os
from ecflow import (
    Defs,
    Suite,
    Family,
    Task,
    Edit,
    Trigger,
    Complete,
    Event,
    Meter,
    Time,
    Day,
    Date,
    Edit,
)


def create_family_f2():
    return Family(
        "f2",
        Edit(SLEEP=20),
        Task("t1", Trigger(":ECF_DATE ==20200720 and :TIME >= 1000")),
        Task("t2", Trigger(":DOW == 4 and :TIME >= 1300")),
        Task("t3", Trigger(":DD == 1 and :TIME >= 1200")),
        Task(
            "t4",
            Trigger("(:DOW == 1 and :TIME >= 1300) or (:DOW == 5 and :TIME >= 1000)"),
        ),
        Task("t5", Trigger(":TIME == 0002")),
    )


print("Creating suite definition")
home = os.path.join(os.getenv("HOME"), "course")
defs = Defs(Suite("test", Edit(ECF_INCLUDE=home, ECF_HOME=home), create_family_f2()))
print(defs)

print("Checking job creation: .ecf -> .job0")
print(defs.check_job_creation())

print("Checking trigger expressions")
errors = defs.check()
assert len(errors) == 0, errors

print("Saving definition to file 'test.def'")
defs.save_as_defs("test.def")



import os
from ecflow import Defs, Suite, Family, Task, Edit, Trigger, Complete, Event, Meter


def create_family_f1():
    return Family(
        "f1",
        Edit(SLEEP=20),
        Task("t1", Meter("progress", 1, 100, 90)),
        Task("t2", Trigger("t1 == complete"), Event("a"), Event("b")),
        Task("t3", Trigger("t2:a")),
        Task("t4", Trigger("t2 == complete"), Complete("t2:b")),
        Task("t5", Trigger("t1:progress ge 30")),
        Task("t6", Trigger("t1:progress ge 60")),
        Task("t7", Trigger("t1:progress ge 90")),
    )


print("Creating suite definition")
home = os.path.join(os.getenv("HOME"), "course")
defs = Defs(Suite("test", Edit(ECF_INCLUDE=home, ECF_HOME=home), create_family_f1()))
print(defs)

print("Checking job creation: .ecf -> .job0")
print(defs.check_job_creation())

print("Saving definition to file 'test.def'")
defs.save_as_defs("test.def")



#!/usr/bin/env python
# ////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
# ////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

# -m trace --count -C .
"""an example for a simple python client to help ecFlow scheduling
 + unit test
 + task loading in the suite

one script DO-IT-ALL:
  - play suite + update suite holder aka def-file

  - script as a wrapper for a task

  - download content (passive client)

  - upload statuses (active client)

$manual

  MIRRORING suite from server to server

  test - suspend suite and rerun job if aborted... Thanks

$end
$comment
  a comment in a comment
$end

TODO: check MICRO robust sync
"""
import argparse
import os
import pwd
import sys
import time
import unittest
from time import gmtime, strftime

# ECFLOWP = "/usr/local/apps/ecflow/current"
# ECFLOWC = ECFLOWP + "/bin/ecflow_client "
# sys.path.append(ECFLOWP + "/lib/python2.7/site-packages/ecflow")
ECFLOWC = "ecflow_client "
child = None


def getreq():
    return Family("getreq").add(Task("collectreq"))


def ymd():
    return [Edit("YMD", "29090909"), Label("ymd", "29090909")]


def excepthook(exctype, value, traceback):
    if exctype == KeyboardInterrupt:
        if child:
            child.report("abort", "keyb")
    else:
        sys.__excepthook__(exctype, value, traceback)
        if child:
            child.report("abort", "gen")


def fullname(item):
    return ""


sys.excepthook = excepthook
from ecf import *  # mirror
import ecf
import ecflow


# class Edit(ecf.Variables):    pass

MICRO = "$$"  # keep ecFlow pleased with micro character balance
DEBUG = 1
DEBUG = 0
USER = "emos"
# USER = "map"

sms_status = {
    -1: "unknown",
    0: "unknown",
    1: "suspended",
    2: "complete",
    3: "queued",
    4: "submitted",
    5: "active",
    6: "aborted",
    7: "shutdown",
    8: "halted",
    9: "unknown",
}
sms_type = {
    13: "definition",
    12: "suite",
    11: "family",
    10: "task",
    32: "alias",
}


############################
def usage():
    print(
        sys.argv[0],
        """
      -h # help
      -m <server@port> # mirror path fri
      -t # unit test
      -r # load the task definition in a suite, associated with -p option
      -r # with -m, allow suite loading in destination
      -u usage

can also be used as a module: gen_task, gen_suite

    """,
    )
    for num, val in enumerate(sys.argv):
        print(num, val)
    child.report("abort")
    sys.exit(2)


############################
def get_uid():
    return pwd.getpwnam(get_username()).pw_uid


############################
def gen_task(var=True, load_only=False, kind=None):
    from ecf import Task, Label, Variables

    pwd = os.getcwd()
    name = sys.argv[0]
    if "/" in name:
        name = name.split("/")[-1]
    if "." in name:
        name = name.split(".")[0]
    rec = "00:30"
    cmd = "/home/ma/emos/bin/trimurti emos eurus $ECF_JOB$ $ECF_JOBOUT$"
    cmd = "/usr/local/apps/python/2.7.12-01/bin/python $ECF_JOB$"
    WDIR = pwd
    ODIR = pwd

    load = (
        Label("info", "mirror suites..."),
        Variables(
            ECF_MICRO=MICRO[0],
            ECF_HOME=ODIR,
            ECF_EXTN=".py",
            ECF_JOB_CMD=cmd,
        ),
    )
    if load_only:
        return load
    return Task(name).add(Cron("04:00 20:00 " + rec), If(var, load))


############################
# def stop(msg, num, child=None):
#     print(msg);
#     if child:
#         child.report(msg)
#         if num == 0: child.report("complete")
#     sys.stdout.flush()
#     sys.stderr.flush()
#     sys.exit(num)
def timer(a, b):
    if child is None:
        return
    child.report("timeout")
    child.report("complete")


import signal

signal.signal(signal.SIGALRM, timer)
signal.alarm(180)


############################
class Child(object):
    """facilitate job behaviour with ecFlow server:
    just instanciate and call complete when finished

    obj = Child(); obs.report("complete")

    this does nothing when script is called from command line
    """

    def __init__(self):
        import signal

        env = {
            "ECF_HOST": "$ECF_HOST:$",
            # check can be on a server, child on another
            "ECF_PASS": "$ECF_PASS$",
            "ECF_NAME": "$ECF_NAME$",
            "ECF_PORT": "$ECF_PORT:0$",
            "ECF_TRYNO": "$ECF_TRYNO$",
        }
        self.client = None
        if MICRO[0] in env["ECF_PORT"]:
            print("#MSG: cli mode")
            return
        print("#MSG: will communicate with server...")
        print("#MSG: kill: ssh %s kill -15 %d" % (os.uname()[1], os.getpid()))
        for sig in (
            signal.SIGINT,
            signal.SIGHUP,
            signal.SIGQUIT,
            signal.SIGILL,
            signal.SIGTRAP,
            signal.SIGIOT,
            signal.SIGBUS,
            signal.SIGFPE,
            signal.SIGUSR1,
            signal.SIGUSR2,
            signal.SIGPIPE,
            signal.SIGTERM,
            signal.SIGXCPU,
            signal.SIGPWR,
        ):
            signal.signal(sig, self.signal_handler)
        self.set_client()

    def set_client(self):
        self.client = ecflow.Client()
        if len(MICRO) == 2:
            host = env["ECF_HOST"]
            print(host)
            self.client.set_host_port(host, int(env["ECF_PORT"]))
            self.client.set_child_pid(os.getpid())
            self.client.set_child_path(env["ECF_NAME"])
            self.client.set_child_password(env["ECF_PASS"])
            self.client.set_child_try_no(int(env["ECF_TRYNO"]))
        else:
            host = "$ECF_HOST:$"
            print(host)
            self.client.set_host_port(host, int("$ECF_PORT:0$"))
            self.client.set_child_pid(os.getpid())
            self.client.set_child_path("$ECF_NAME$")
            self.client.set_child_password("$ECF_PASS$")
            self.client.set_child_try_no(int("$ECF_TRYNO$"))
        self.client.child_init()

        self.client.child_label("info", "%s" % strftime("%Y-%m-%d %H:%M:%S", gmtime()))
        self.client.set_child_timeout(20)
        print(self.client)

    def signal_handler(self, signum, frame):
        """ catch signal """
        print("Aborting: Signal handler called with signal ", signum)
        self.report("abort", "Signal handler called with signal " + str(signum))

    def __exit__(self, exc_type, exc_value, traceback):
        self.report("abort", "__exit__")

    def report(self, msg, meter=None):
        """ communicate with ecFlow server """
        if not self.client:
            print("#MSG: no child")
            if msg in ("stop", "complete"):
                sys.stdout.flush()
                sys.stderr.flush()
                print("#MSG: no child stop")
                sys.exit(0)
            else:
                print(msg, meter)
            return
        elif msg in ("stop", "complete"):
            self.client.child_complete()
            self.client = None
            sys.stdout.flush()
            sys.stderr.flush()
            print("#MSG: stop")
            sys.exit(0)
        elif msg in ("abort",):
            self.client.child_abort()
            self.client = None
            raise Exception(msg)
        elif meter:
            self.client.child_meter(msg, meter)
        else:
            self.client.child_label("info", msg)


############################
destinations = {
    "emos": "${DEST_HOST:=localhost}@${DEST_PORT=5001}",
    "test": "${DEST_HOST:=localhost}@${DEST_PORT=5001}",
    "mc": "localhost@5003",
}


def gen_suite(host=None, port=None, path=None):
    import ecf
    from ecf import (
        Task,
        Label,
        Variables,
        Suite,
        Defs,
        Family,
        Trigger,
        Defstatus,
        Meter,
        Event,
    )

    # destinations = dict()
    defs = Defs()
    sname = "mirror"
    print(list(definitions.keys()), list(destinations.keys()))
    defs.add_suite(
        Suite(sname).add(
            ecf.Defstatus("suspended"),
            Limit("one", 2),
            Inlimit("one"),
            gen_task(load_only=True),
            Variables(ALL_ECF=" "),
            [
                Family(name).add(
                    Variables(DESTINATIONS=destinations[name]),
                    Label("info", ""),
                    gen_task(0, kind=name),
                )
                for name in list(definitions.keys())
                # if name in destinations.keys()
            ],
        )
    )
    if DEBUG:
        print(defs)
    return defs


definitions = {  # strings as path for all nodes below sync
    # or suites, subset of the original, leaves are task...
    # or else not sync'ed
    # shall we add list of string/path ???
    "emos": Suite("emos").add(
        Defstatus("suspended"), Family("f1").add(Edit(YMD=20320101), Task("t1"))
    ),
    "test": Suite("test").add(
        Defstatus("suspended"), Family("f1").add(Edit(YMD=20320101), Task("t1"))
    ),
    "mc": Suite("mc").add(
        Defstatus("suspended"), Family("main").add(
            Edit(YMD=20990101), 
            Family("06bc").add(
                Task("sweeper"), 
                Family("fc0006d").add(Task("fc")),
            ),
    ), ),
}

############################


class Mirror(object):

    """    """

    def _protect_oper(self, server):
        if server == "xyz@123":
            if 1:
                child.report("abort", "protect")
            elif (
                "/o" in self.path
                or self.path == "all"
                or "/eda" in self.path
                or "/mc" in self.path
            ):
                raise Exception("# wont do that...")

    def _warmup(self):
        self._set_register(self.path)
        if not self.is_sms:
            self.source.ch_register(False, self.register)
            self.source.sync_local()

    def _send(self, msg="", client=None, send=False):
        """self.grouped is None: no buffering, immediate send
        msg accumulate (send False by default)
        msg is sent ("", client, send=True)
        """
        if self.grouped is None:
            print("#MSG: send", self.grouped)
            client.group(msg)
            time.sleep(self.snap)
            return

        self.grouped += msg
        if send and self.grouped != "":
            print("#MSG: send", self.grouped)
            client.group(self.grouped)
            client.sync_local()
            self.grouped = ""
            time.sleep(self.snap)

    def __init__(self, server="eurus@1630", path="/e_45r1", replay=False, is_sms=False):
        self.is_sms = is_sms
        self.replay = replay  # allow replay from skel definition
        self.act = 1  # active ie force change allowed
        self.snap = 1  # sleep interval
        self.wait = 0  # refresh interval,continue as long as sub/act node exist
        self.path = path
        self.grouped = None  # would use atomic alter
        self.grouped = ""  # command to send to the server ... grouped
        self._set_register(path)

        if server == "all":
            self.servers = "$DESTINATIONS$".split(" ")
        elif " " in server:
            self.servers = server.split(" ")
        elif "@" in server:
            host, port = server.split("@")
            self.servers = ("%s@%s" % (host, port),)
        elif ":" in server:
            host, port = server.split(":")
            self.servers = ("%s:%s" % (host, port),)
        if self.servers[-1] == "":
            del self.servers[-1]
        host = os.getenv("ECF_HOST", "localhost")
        # if MICRO[0] in host: host = "localhost"
        port = os.getenv("ECF_PORT", "3141")
        if MICRO[0] in port:
            port = 1500 + int(get_uid())
        # when started with ecflow_start.sh
        # aka reference/source  server name
        self.sname = "%s:%s" % (host, port)

        if "$SOURCE:$" != "" and MICRO[0] not in "$SOURCE:$":
            self.sname = "$SOURCE:$"
        if DEBUG:
            print("#DBG: source is", self.sname)
        self.set_source()

    def set_source(self):
        self.source = ecflow.Client(self.sname)
        self._warmup()

    def __exit__(self):
        if type(self.source) == ecflow.Client:
            try:
                self.source.ch_drop()
            except:
                pass

    def _set_register(self, path):
        if "/" == path[0]:
            suite = str(path.split("/")[1])
            self.register = [
                suite,
            ]

        elif path == "all":
            self.register = list(skel.keys())

        else:
            child.report("abort", "register")

    def _check(self, server, client, stop=0):
        if client == self.source:
            stop = 1

        for suite in self.register:
            top = client.get_defs().find_abs_node("/" + suite)

            if top is None:
                print(
                    "#WAR: node not found",
                    suite,
                    server,
                )
                if stop:
                    child.report("abort", "check")
                elif self.act and server != self.sname:
                    self._protect_oper(server)
                    if self.replay:
                        if type(definitions[suite]) == str:
                            print("#WAR: not yet")
                            child.report("abort", "str")
                        print("replaying")

                        defs = ecflow.Defs()
                        defs.add_suite(definitions[suite])
                        host = "localhost"
                        client.load(defs)
                        client.suspend("/" + suite)
                        client.begin_suite(suite)
                        client.sync_local()
                    else:
                        print("replay?")
                else:
                    pass

            elif server != self.sname and not (
                top.is_suspended()
                or "%s" % top.get_state() == "unknown"  # normal backup server
            ):  # test server
                print("#ERR: top node shall be suspended - NOGO")
                print("#ERR:", server, top.name(), top.get_state())
                if 1:
                    child.report("abort", "state")
                print("#MSG: stop")
                child.report("abort", "state")  # sys.exit(1)

            else:
                pass

            if suite not in list(definitions.keys()):
                raise Exception(list(definitions.keys()), suite)

            if type(definitions[suite]) == str:
                path = definitions[suite]

            elif suite in self.path:
                path = self.path

            else:
                return

            node = client.get_defs().find_abs_node(path)
            if node:
                return
            print("#WAR: node not found", path, server)
            if stop:
                child.report("abort", "stop")

    def _update_var(self, var, nfrom, ndest, server, client, kind="edit"):
        name = var.name()
        vfrom = None
        vdest = None
        repeat = nfrom.get_repeat()
        print("############# %s" % repeat, nfrom.get_abs_node_path())
        if repeat and repeat.name() == name:
            vfrom = repeat
        else:
            for vfrom in nfrom.variables:
                if vfrom.name() == name:
                    break

        for vdest in ndest.variables:
            if vdest.name() == name:
                break
        pnode = ndest.get_abs_node_path()
        pnode = nfrom.get_abs_node_path()

        if vfrom is None:
            return

        elif vfrom.name() != name:
            return

        if vdest is None:
            return

        vname = vdest.name()
        vval = vdest.value()
        if vname != name:
            return

        if vfrom.name() == vname and str(vfrom.value()) != str(vval) or FORCE:
            if type(vfrom) == Repeat:
                self.action(
                    ndest.get_abs_node_path(),
                    server,
                    client,
                    kind="repeat",
                    value=str(vfrom.value()),
                    msg="repeat requeue",
                )
            else:
                self.action(
                    pnode,
                    server,
                    client,
                    kind=name,
                    value=str(vfrom.value()),
                    msg="value update",
                )

    def process(self, server=None, client=None, item=None, var=False):
        if server is None:
            self._check(self.sname, self.source, 1)

            for server in self.servers:
                server = server.replace("@", ":")
                if DEBUG:
                    print("#DBG: target is", server)
                host, port = server.split(":")

                if int(port) < 65536 and 0 == comm(
                    ECFLOWC + "--ping --port %s --host %s" % (port, host), False
                ):
                    client = ecflow.Client(server)
                    client.ch_register(False, self.register)
                    client.sync_local()
                    self._check(server, client)
                    for suite in self.register:
                        node = definitions[suite]
                        if type(node) == str:
                            node = self.source.get_defs().find_abs_node(node)
                        # try: node.add(Label("memo", "This suite was generated by mirror.py"))
                        # except: pass # used to sync, too late
                        # update variables
                        self.process(server, client, node, "repeat")
                        # self.process(server, client, node) # update statuses
                        # self.process(server, client, node, "edit") # update
                        # variables
                    client.ch_drop()

                else:
                    # child.report("abort","res")
                    print("server is not responding")

            if self.wait > 0:
                print("#MSG: start again... snap ", self.wait)
                time.sleep(self.wait)
                self.wait = 0
                self.source.sync_local()
                self.process()
            return

        elif server == self.sname:
            child.report("abort", "same name " + server + " " + self.sname)  # DONT

        elif client is None:
            child.report("abort", "no client")  # DONT

        elif type(item) in (ecflow.Suite, ecf.Suite):

            if 0:
                pass
            else:
                # suite =
                # client.get_defs().find_abs_node(item.get_abs_node_path())
                suite = client.get_defs().find_abs_node(item.fullname())
                # print(type(client), item.fullname(), type(suite))
                if suite:
                    if not suite.is_suspended():
                        print("#ERR: top node shall be suspended - stop", suite, server)
                        return
                    for node in suite.nodes:
                        self.process(server, client, node, var)
            return

        elif type(item) in (ecflow.Family, ecflow.Task):
            pnode = item.get_abs_node_path()
            nfrom = self.source.get_defs().find_abs_node(pnode)
            if 0:
                pass
            else:
                ndest = client.get_defs().find_abs_node(pnode)

            if nfrom is None:
                if var:
                    return  # ignore
                print(
                    "#ERR: node not found, please update mirror.py skel",
                    pnode,
                    self.sname,
                )
                child.report("abort", "nfrom")

            elif ndest is None:
                if var:
                    return  # ignore
                print("#ERR: node not found, replay?", pnode, server)
                return

            else:
                pass
            if nfrom is not None:
                status = "%s" % nfrom.get_state()
            else:
                raise Exception(pnode, self.source, "nfrom is not")

            if 0:
                pass
            else:
                dstate = "%s" % ndest.get_state()
            is_task = type(item) == ecf.Task

            for evs in nfrom.events:
                for evd in ndest.events:
                    # if evd.name_or_number() != evs.name_or_number(): continue
                    if evd.name() == "" and evd.number() != evs.number():
                        continue
                    if evd.value() == evs.value():
                        break
                    print(
                        "#event",
                        evs.name(),
                        evd.number(),
                        evs.number(),
                        evd.value(),
                        evs.value(),
                    )
                    if evs.value():
                        if evs.name() == "":
                            self._send(
                                "alter change event %d set %s;" % (evs.number(), pnode),
                                client,
                            )
                            # client.alter(pnode, "change", "event", "%d" %
                            # evs.number(), "set")
                        else:
                            self._send(
                                "alter change event %s set %s;" % (evs.name(), pnode),
                                client,
                            )
                        # else: client.alter(pnode, "change", "event",
                        # evs.name(), "set")

            for evs in nfrom.meters:
                for evd in ndest.meters:
                    if evd.name() != evs.name():
                        continue
                    if evd.value() == evs.value():
                        break
                    print("#meter", evd.value(), evs.value())
                    try:
                        value = evs.value()
                        if value > 100:
                            value = 100
                        if value > 0:
                            self._send(
                                "alter change meter %s %s %s;"
                                % (evs.name(), value, pnode),
                                client,
                            )
                        # client.alter(pnode, "change", "meter", evs.name(),
                        # "%s" % evs.value())
                    except:
                        pass

            if status != dstate or FORCE:
                print(
                    "#WAR: diff ",
                    pnode,
                    status,
                    dstate,
                )
                if status == "complete":
                    self.action(
                        pnode, server, client, status, is_task, msg="forced complete"
                    )
                    if not is_task:
                        self._send("", client, send=True)
                        return

                elif status == "queued":
                    # self._send("force queued %s;" % (pnode), client)
                    self.action(
                        pnode, server, client, status, is_task, msg="forced queued"
                    )
                elif status in (
                    "submitted",
                    "active",
                    "unknown",
                    "aborted",
                ):
                    print("#IGN: status is active/submit, ignored")
                else:
                    raise Exception(status)

            if type(item) in (ecflow.Suite, ecflow.Family):
                for node in item.nodes:
                    self.process(server, client, node, var)

            if var:
                for edit in item.variables:
                    self._update_var(edit, nfrom, ndest, server, client, var)

            self._send("", client, send=True)
        else:
            raise Exception(type(item), item)

    def action(
        self, pnode, server, client, kind, is_task=False, value="20010101", msg=""
    ):
        if not self.act:
            return

        self._protect_oper(server)

        print("#MSG:", msg)
        if kind in (
            "active",
            "submitted",
        ):
            return
        elif FORCE:
            pass  # SMS ONLY

        elif kind == "complete":
            if is_task:
                self._send("", client, True)
                client.force_state(pnode, ecflow.State.complete)
            else:
                self._send("", client, True)
                client.force_state_recursive(pnode, ecflow.State.complete)
        elif kind == "queued":  # and is_task:
            self._send("force queued %s;" % pnode, client)

        elif kind == "repeat":
            raise Exception

        else:  # var
            if DEBUG:
                print("#DBG: update var", kind, value)
            if kind in ("queued",):
                pass
            elif kind not in (
                "YMD",
                "JUL",
                "DAY",
                "SCHOST",
                "STHOST",
                "YYYY",
                "MM",
                "ECF_FILES",
                "INIBEGINDATE",
                "MIXTASK",
            ):
                raise Exception(kind)
            if 0:
                pass
            else:
                client.requeue(pnode)
                self._send(
                    "alter change variable %s %s %s;" % (kind, value, pnode),
                    client,
                    True,
                )


# ignore Family + queued while is might just be that we dont mirror all
# fam/tasks and inheritance get it already complete in dest while still
# queued in source


############################
def replay(path, defs=None):
    import ecf

    miss = "@undef@"
    host = os.getenv("ECF_HOST", miss)
    if host == miss:
        print("ECF_HOST not defined")
        exit(1)
    port = os.getenv("ECF_PORT", 31415)
    client = ecf.Client(host, port)
    if defs is None:
        defs = gen_suite(host, port, path)
    if DEBUG:
        print(defs)
    # print(defs)
    if (
        "localhost" in host
        and port in ("3141", 3141)
        and ("/test" in path)
        and "/mirror" not in path
    ):
        raise Exception("dont")

    print("#MSG: replacing %s in " % path, host, port, defs)
    client.replace(path, defs, 1, 1)
    return client


def comm(cmd, rem=True):
    import subprocess

    if rem and cmd in list(memo.keys()):
        return memo[cmd]
    (rc, res) = subprocess.getstatusoutput(cmd)
    print("#COMM:", rc, res, cmd)
    if rem:
        memo[cmd] = rc
    return rc


############################
class TestMirror(unittest.TestCase):
    """ a test case """

    def test_1(self, test_ok=1):
        sname = "mirror"
        replay("/" + sname, gen_suite())


memo = dict()
############################
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("-?", default=0, help="help")
    parser.add_argument("-f", "--force", action="store_true", help="force")
    parser.add_argument("-h", default=0, help="help")
    parser.add_argument("-m", "--mirror", default="localhost@31415", help="mirror")
    parser.add_argument("-p", "--path", default="/mirror/test", help="path")
    parser.add_argument("-r", "--replay", action="store_true", help="replay")
    parser.add_argument("-t", "--test", action="store_true", help="test")
    parser.add_argument("-u", "--usage", action="store_true", help="test")

    parsed = parser.parse_args()
    global FORCE
    FORCE = parsed.force
    print(parsed.path, parsed.mirror, parsed.replay, parsed.force, parsed.test)
    # raise Exception
    child = Child()

    if parsed.usage:
        print(
            """

* mirrors (slow real-time) a source suite, from a source server, to a destination server
* mirror suite can be hosted on the source, on the destination or another server
* this script was used in ecflow course to demonstrate
  * one script that can be used from the command line
  * defines a suite
  * play/replace a suite/node
  * a script can be used as task wrapper/template to generate jobs
  * an ecflow client (passive) to download status/value from source
  * an ecflow client (active) to change status/value on destination

* update mirror.py: destination variable with the triplet suite host port

* update mirror.py: definition with the key(suite) and the minimum suite-tree to mirror

* update mirror.py: replay add "if" for "protection" not to overwrite source suite by mistake

* download a copy from the source suite
  ecflow_client --port 3141 --host localhost --get > tmp.def

* load the mirror suite on the source server
  SUITE=test ECF_HOST=localhost ECF_PORT=3141 ./mirror.py -p /mirror/test -r # load

* play (suspended) mirrored suite in destination
  ECF_HOST=ecgate ECF_PORT=31415 ./mirror.py -m ecgate:31415 -p /test

module load python3; module load ecflow/5.7.0
ecflow_start.sh -p 5001 # source
ecflow_start.sh -p 5003 # destination
# ecflow_client --host vecf1 --port 44444 --get mc > mc.dump # an example suite
ecflow_client --host localhost --port 5001 --replace /mc mc.dump # we take it as a source
# definitions =  # contains "mc" definition
# destinations = { # contains one destination for mc at least
 export  SUITE=mc ECF_HOST=localhost ECF_PORT=5001
 ./mirror.py -p /mirror/mc -r # load mirror on source server # might be destination or another server yet
# ImportError: No module named ecf # ecf shall be in the path, at least it can found in ecflow distrib as 
# ecflow/libs/pyext/samples/api/ecf.py
# ('#MSG: replacing /mirror/test in ', 'localhost', '5001', #5.7.0 # loading OK
# begin mirror - 
# Script OK? mirror.py undef ECF_FILES directory?
# Edit OK? $ is used as ECF_MICRO
# cron OK?
# set DESTINATIONS variable as expected

ECF_HOST=localhost ECF_PORT=5003 python3 ./mirror.py -m localhost:5003 -p /mc # ok
#COMM: 0 ping server(localhost:5003) succeeded in 00:00:00.001058  ~1 milliseconds ecflow_client --ping --port 5003 --host localhost # OK
# replay?
ECF_PORT=5003 python3 ./mirror.py -m localhost:5003 -p /mc -r # ok mc loaded on target
# begin /mc in 5003 (it looks suspended
# edit ECF_JOB_CMD "python3 $ECF_JOB$ > $ECF_JOBOUT$ &"

# python3 /home/ma/map/mirror/mc/mirror.job1 # OK after ECF_PASS FREE running as zombie ommand line
# ImportError: No module named 'ecf' # jobs lost in submit

# mirror jobs submits and complete OKWITH ECF_JOB_CMD updated as 
module load python3 && module load ecflow/5.7.0 && cd /home/ma/map && python3 $ECF_JOB$ > $ECF_JOBOUT$ 2>&1 &
"""
        )
        sys.exit(2)

    if not MICRO[0] in "$ECF_PORT$":
        parsed.mirror = "$DESTINATIONS$"
        parsed.path = "/$FAMILY1$"
        print("#MSG: ssh %s kill -9 %d" % (os.uname()[1], os.getpid()))

    if parsed.test:
        sys.argv.pop()
        unittest.main()

    elif parsed.path and parsed.replay:
        print("#MSG: path and replay", parsed.path, parsed.replay)
        port = int(os.getenv("SMS_PROG", 0))
        s = parsed.path.split("/")[1]
        defs = None
        if "mirror" not in parsed.path:
            defs = Defs()
            defs.add_suite(definitions[s])
        replay(parsed.path, defs)

    elif parsed.mirror and parsed.path:
        print(parsed.mirror, "#", parsed.path, parsed.replay, type(Mirror))
        mirror = Mirror(parsed.mirror, parsed.path, parsed.replay)
        print("#MSG: step")
        mirror.process()
        print("#MSG: step")
        child.report("complete")
        print("#MSG: step")

    else:
        usage()
    print("#MSG: step")
    child.report("complete")
"""
  export ECF_HOST=localhost ECF_PORT=5001 SUITE=mc
  ecflow_client --get /mc > tmp.def
  ./mirror.py -p /mirror/mc -r # load

  ECF_HOST=localhost ECF_PORT=31415 ./mirror.py -m localhost:31415 -p /mc

/tmp/map/work/git/ecflow/libs/pyext/samples/mirror.py
"""



#!/usr/bin/env python
# -*- coding= UTF-8 -*-
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# file deepcode ignore MissingCloseOnSomePath: <comment the reason here>

import sys, os, pwd, getopt
sys.path.append('/home/ma/emos/def/o/def')
import inc_emos as ic
from ecf import *
import ecf
from inc_emos import Family, Task, Variables, Time, Meter, Event, Label, Late
import sms2ecf

# ecf.USE_TRIGGER = 0 # DEBUG MODE: do not load triggers

def get_username():
    return pwd.getpwuid( os.getuid() )[ 0 ]
def get_uid():
    return pwd.getpwnam(get_username()).pw_uid

def create_wrapper(name, content):
    print("#MSG: creating file %s/" % wdir + name)
    wrapper = open(wdir + "/%s" % name, 'w')
    print(content, file=wrapper)
    wrapper.close()    
##############################################

# ic.SUBM = "/home/ma/emos/bin/smssubmit.cray"
# ic.KILL = "/home/ma/emos/bin/smssubmit.cray"
# ic.STAT = "/home/ma/emos/bin/smssubmit.cray"
# sms2ecf.subm = "/home/ma/emos/bin/smssubmit.cray"
# sms2ecf.kill = "/home/ma/emos/bin/smssubmit.cray"
# sms2ecf.stat = "/home/ma/emos/bin/smssubmit.cray"

def create_task():
    ### PRODUCER/CONSUMER TASK
    produce = '''#!/bin/ksh
%manual
  start - check - stop logsvr
%end
%include <qsub.h>
%include <trap.h>

step=%BEG:0%
while [ $step -le %FIN:48% ]; do
  if [[ %PRODUCE:1% = yes ]]; then
    xevent p
  elif [[ %CALL_WAITER:0% != 0 ]]; then
case %ECF_PROG:0% in 
0) smswait %TRIGGER:1==1%;;
*) ecflow_client --wait "%TRIGGER:1==1%";;
esac
  fi

  if [[ %CONSUME:no% = yes ]]; then
    xevent c
  fi

  xmeter step $step
  step=$((step + %BY:1%))
done
 
%include <endt.h>
'''
    create_wrapper("produce.sms", produce)

    create_wrapper("consume.sms", produce)
    ### PURE PYTHON

    head = """import os, time, signal
import ecflow
  
print "PYTHONPATH====================================================="
print os.environ['PYTHONPATH'].split(os.pathsep)
 
class Client(object):
    ''' communication with the ecflow server. This will automatically call
       the child command init()/complete(), for job start/finish. It will also
       handle exceptions and signals, by calling the abort child command.
       *ONLY* one instance of this class, should be used. Otherwise zombies will be created.
    '''
    def __init__(self):
      print "Creating Client"
      self.ci = ecflow.Client()
      self.ci.set_host_port("$ECF_HOST$","$ECF_PORT$")
      self.ci.set_child_pid(os.getpid())
      self.ci.set_child_path("$ECF_NAME$")
      self.ci.set_child_password("$ECF_PASS$")
      self.ci.set_child_try_no($ECF_TRYNO$)
    
      print "Only wait 20 seconds, if the server cannot be contacted (note default is 24 hours) before failing"
      self.ci.set_child_timeout(20)
      self.ci.set_zombie_child_timeout(10)
     
      # Abort the task for the following signals
      signal.signal(signal.SIGINT,  self.signal_handler)
      signal.signal(signal.SIGHUP,  self.signal_handler)
      signal.signal(signal.SIGQUIT, self.signal_handler)
      signal.signal(signal.SIGILL,  self.signal_handler)
      signal.signal(signal.SIGTRAP, self.signal_handler)
      signal.signal(signal.SIGIOT,  self.signal_handler)
      signal.signal(signal.SIGBUS,  self.signal_handler)
      signal.signal(signal.SIGFPE,  self.signal_handler)
      signal.signal(signal.SIGUSR1, self.signal_handler)
      signal.signal(signal.SIGUSR2, self.signal_handler)
      signal.signal(signal.SIGPIPE, self.signal_handler)
      signal.signal(signal.SIGTERM, self.signal_handler)
      signal.signal(signal.SIGXCPU, self.signal_handler)
      signal.signal(signal.SIGPWR,  self.signal_handler)
     
    def signal_handler(self,signum, frame):
      print 'Aborting: Signal handler called with signal ', signum
      self.ci.child_abort("Signal handler called with signal " + str(signum));
     
    def __enter__(self):
      self.ci.child_init()
      return self.ci
     
    def __exit__(self,ex_type,value,tb):
      print "Client:__exit__: ex_type:" + str(ex_type) + " value:" + str(value) + "\\\n" + str(tb)
      if ex_type is not None:
         self.ci.child_abort("Aborted with exception type " + str(ex_type) + ":" + str(value))
         return False
      self.ci.child_complete()
      return False
"""
    body = """#!/usr/bin/env python
$include <head.py>
 
if __name__ == "__main__":
# This will also handle call to sys.exit(), i.e Client.__exit__ will still be called.
   with Client() as ci:
   # *******************************************************************************
   # This is where the main work is done.
   # *******************************************************************************
      for i in range(1,100):
         ci.child_meter('step',i)
         ci.child_label('info', "value_" + str(i))
         time.sleep(1)
       
      ci.child_event('1')
      print "Finished event,meter and label child commands"
 
$manual
   This is the manual section. Instead of calling python from the ECF_JOB_CMD we could alternatively place,
       #!/bin/env/python
   on the first line of this file.
$end
 
$comment
   Note: We do not need a include a tail.py, the head.py does it all.
$end

"""
    create_wrapper("head.py", head)
    create_wrapper("pure_python.sms", body)
    ### PYTHON TASK
    content = """$include <python_header.h>
# header files are located in the same directory as wrapper (quotes)
for step in range(0,101):
    print step
    xmeter("step", step)
else:
    print 'the loop is over'
xevent("1")
xlabel("info", "news from pure python world")

$include <python_endt.h>
"""
    create_wrapper("python.sms", content)

    content = """#!/usr/bin/env python
import os
import sys
import signal 

ECF_PORT=$ECF_PORT:0$
XECF="/usr/local/apps/ecflow/current/bin/ecflow_client ";
# --port=$ECF_PORT:0$ --host=$ECF_HOST:0$ ";
def SigHandler(signum, frame):
   print "caught signal " + signum
   xabort()
   sys.exit(0);
   return

print ECF_PORT
# set -eux ?
# trap 0 ?
# time stamp per executed line

import atexit
early_exit = True
@atexit.register
def goodbye():
  if early_exit: 
    print "too early"
    xabort()
  else: 
    xcomplete()

# TIME_STAMP
# http://shop.oreilly.com/product/9780596007973.do # recipie p436
import syslog, time

class FunctionFileLikeWrapper():
    def __init__(self, func): self.func = func
    def write(self, msg): self.func(msg)
    def flush(self): pass

class TimeStamper(object):
    msg_format = "[%y%m%d %H:%M:%S]", time.gmtime, "%s: %s"
    msg_format = "+ %H:%M:%S", time.gmtime, "%s %s"

    def __call__(self, msg):
        tfmt, tfun, gfmt = self.msg_format
        return "%s %s\\\n" % (time.strftime(tfmt, tfun()), msg)

class TeeFileLikeWrapper():
    def __init__(self, *files): self.files = files
    def write(self, msg): 
        for f in self.files: f.write(timestamp(msg.strip()))

class FlushingWrapper:
    def __init__(self, *files): self.files = files
    def write(self, msg):
        for f in self.files:
            f.write(timestamp(msg))
            # f.write(timestamp(msg.strip()))
            f.flush()

def logto(*files):
    # sys.stdout = TeeFileLikeWrapper(*files)
    sys.stdout = FlushingWrapper(*files)

syslogger = syslog.syslog
syslogfile = FunctionFileLikeWrapper(syslogger)
timestamp = TimeStamper()
logto(sys.stdout, syslogfile, open("log.tmp", "w"))
# end time stamp

if ECF_PORT > 0:
  os.environ['ECF_PORT'] = "$ECF_PORT:0$"
  os.environ['ECF_NAME'] = "$ECF_NAME:0$"
  os.environ['ECF_HOST'] = "$ECF_HOST:0$"
  os.environ['ECF_PASS'] = "$ECF_PASS:0$"

  def xinit():
    os.system(XECF + " --init=" + str(os.getpid()))  
    print "init"
  def xabort():
    os.system(XECF + " --abort")  
  def xcomplete():
    os.system(XECF + " --complete")  
  def xmeter(name, step):
    os.system(XECF + " --meter=" + name + " " + str(step))  
  def xevent(name):
    os.system(XECF + " --event=" + name)  
  def xlabel(name, msg):
    os.system(XECF + " --label=" + name + " '%s'" % msg)  
else:
  os.environ['SMS_PROG'] = "$SMS_PROG:0$"
  os.environ['SMSNAME'] = "$SMSNAME:0$"
  os.environ['SMSNODE'] = "$SMSNODE:0$"
  os.environ['SMSPASS'] = "$SMSPASS:0$"

  def xinit():
    os.system('smsinit ' + str(os.getpid()))
  def xabort():
    os.system('smsabort')
  def xcomplete():
    os.system('smscomplete')
  def xmeter(name, step):
    os.system("smsmeter " + name + " " + str(step))
  def xevent(name):
    os.system("smsevent " + name)
  def xlabel(name, msg):
    os.system("smslabel " + name + " '%s'" % msg)

signal.signal (signal.SIGHUP,  SigHandler)
signal.signal (signal.SIGINT,  SigHandler)
signal.signal (signal.SIGQUIT, SigHandler)
signal.signal (signal.SIGILL,  SigHandler)
signal.signal (signal.SIGTRAP, SigHandler)
signal.signal (signal.SIGIOT,  SigHandler)
signal.signal (signal.SIGBUS,  SigHandler)
signal.signal (signal.SIGFPE,  SigHandler)

"""
    create_wrapper("python_header.h", content)

    content = """# os.system('smscomplete')
early_exit = False
# xcomplete() # managed with atexit
# os.system('/usr/local/apps/sms/bin/ecflow/bin/ecf_client --complete')
"""
    create_wrapper("python_endt.h", content)

    ### PERL TASK
    create_wrapper("perl.sms", """#!/usr/bin/perl -w
^include <perl_header.h>
# header files are located in the same directory as wrapper (quotes)
print "Pure perl SMS task";
for ( my $step=1; $step <= 100 ; $step++ ) {
   print "this is the number $step\\\n";
   xmeter("step", $step);
}
xevent("1");
            
xlabel("info", "news from pure perl world");      
^include <perl_endt.h>
""")
    # create_wrapper("perl.sms", content)

    source = """use strict;

my $xmeter = "smsmeter"; my $arg_m = "";
my $xlabel = "smslabel"; my $arg_l = "";
my $xevent = "smsevent"; my $arg_e = "";
my $xcomplete = "smscomplete"; my $arg_c = "";
my $xabort = "smsabort";

if (^ECF_PORT:0^ != 0) {
$ENV{'ECF_PORT'}  = "^ECF_PORT:0^" ;  # ecFlow port number
$ENV{'ECF_HOST'}  = "^ECF_HOST:0^"  ; # ecFlow host
$ENV{'ECF_NAME'}  = "^ECF_NAME:0^"  ; # task path into the suite
$ENV{'ECF_PASS'}  = "^ECF_PASS:0^"  ; # password for the job
$ENV{'ECF_TRYNO'} = "^ECF_TRYNO:0^" ; # job occurence number
my $client = "/usr/local/apps/ecflow/current/bin/ecflow_client";
$xmeter = $client; $arg_m = "--meter";
$xlabel = $client; $arg_l = "--label";
$xevent = $client; $arg_e = "--event";
$xcomplete = $client; $arg_c = "--complete";
$xabort = $client; 

system($client, "--init", "$$");
} else {
  $ENV{'SMS_PROG'} = "^SMS_PROG:0^" ; # SMS Program Number
  $ENV{'SMSNODE'}  = "^SMSNODE:0^"  ; # SMS host
  $ENV{'SMSNAME'}  = "^SMSNAME:0^"  ; # task path into the suite
  $ENV{'SMSPASS'}  = "^SMSPASS:0^"  ; # password for the job occurence
  $ENV{'SMSTRYNO'} = "^SMSTRYNO:0^" ; # job occurence number
}

sub xmeter($$){ my ($name, $step) = @_; 
  system($xmeter, $arg_m, $name, $step); }
sub xevent($){ my ($name) = @_; 
  system($xevent, $arg_e, $name); }
sub xlabel($$){ my ($name, $msg) = @_; 
  system($xlabel, $arg_l, $name, $msg); }
sub xabort(){ system($xabort); }
sub xcomplete(){ system($xcomplete, $arg_c, "$$"); }

print "start";
eval '
"""
    create_wrapper("perl_header.h", source)

    source="""';
if ($@){
    print "caught signal: $@\n";
    xabort();
    exit;
  }
print "the job is now complete\n";
xcomplete();
exit;"""
    create_wrapper("perl_endt.h", source)

    ### LOGER TASK
    content = """#!/bin/ksh
%manual
  start - check - stop logsvr
%end
%include <qsub.h>
%include <trap.h>
SLEEP=%SLEEP:0%
KIND=%KIND:start%
type rsh && RSH=rsh || RSH=ssh

case $HOST in 
c2a*) rhost=c2a;;
c2b*) rhost=c2b;;
xxcct*) rhost=cctdtn1 ; RSH=ssh;;
cct*) rhost=cct-log ; RSH=ssh;;
cca*) rhost=cca-log ; RSH=ssh ;;
xxcca*) rhost=cca-il2 ; RSH=ssh ;;
xxccb*) rhost=ccb-il2 ; RSH=ssh ;;
ccb*) rhost=ccb-log ; RSH=ssh ;;
esac

case $KIND in
start)

ps -elf | grep logsvr | grep -v grep | grep $USER && \\\
ps -elf | grep logsvr | grep $USER | grep -v grep | \\\
awk '{print $1}' | xargs kill -9

case $ARCH in
cray)  
  case $HOST in
  cct*) logsvr=/usr/local/apps/sms/bin/logsvr.sh
        # /usr/local/apps/emos/bin/logsvr.sh
        # $RSH $rhost 
        nohup $logsvr > /tmp/emos_logsvr.tmp 2>&1 &
        rhost=cct # start on both nodes cctdtn1  + cct (Avi serial jobs)
        trap 0; xcomplete; exit 0
  ;;
  cca*|ccb*) logsvr=/usr/local/apps/emos/bin/logsvr.sh;;
  *)    exit 1;;
  esac

  $RSH $rhost nohup $logsvr > /tmp/emos_logsvr.tmp 2>&1 &
;;
ibm*) $RSH $rhost nohup $HOME/bin/boot/logsvr.sh > /tmp/emos_logsvr.tmp 2>&1 &
;;
*)    nohup logsvr.sh > /tmp/emos_logsvr.tmp 2>&1 &
;;
esac
xlabel info "logserver is started"

;;
stop)
pid=$($RSH $rhost ps -elf | grep logsvr | grep $USER | grep -v grep | awk '{print $1}' )
kill -9 $pid
xlabel info "logserver is stoped"
;;

ping)
case $ARCH in
cray) ps -elf | grep logsvr | grep $USER # grep -v grep || exit 1
      xlabel info "???"
     trap 0; xcomplete; exit 0
;;
*) $RSH $rhost ps -elf | grep logsvr | grep -v grep | grep $USER || exit 1 
;;
esac

xlabel info "$($RSH $rhost ps -elf | grep logsvr | grep -v grep | grep $USER)"
;;
esac

%include <endt.h>
"""
    create_wrapper("logsvr.sms", content)

    ### task
    content = """#!/bin/ksh
%manual 
manual - this task is automatically created by cray.py
%end
%include <qsub.h>
if [[ $ARCH = hp* ]]; then export PATH=/usr/local/bin:$PATH; fi
if [[ $HOST = lxop* ]]; then export PATH=/usr/local/apps/ecflow/current:/usr/local/apps/sms/bin:$PATH; fi
%include <trap.h>
SLEEP=%SLEEP:0%
echo OK

case %ECF_PORT:0% in
0) base=900000; LOGPORT=%LOGPORT:0%;;
*) base=1000; LOGPORT=%LOGPORT:0%;;
esac
base=1000

case $HOST in
cc*)xlabel info $(printenv | grep -E '(SUBMIT_|EC_)') ;;
*) xlabel info OK
esac
  
printenv | sort
xevent 1
step=0
while (( $step <= 12 )); do
  xmeter step $step
  ((step = step + 1))
  sleep $SLEEP
done
%include <endt.h>
"""
    create_wrapper("test.sms", content)

    content = """#!/bin/ksh
%manual 
manual - this task is automatically created by cray.py

expected to run in real-time mode only

%end
#include <qsub.h>
#include <step1.h>
%include <trap.h>

KIND=%KIND:when%
NOW=$(date +%%H%%M)

case $KIND in 
when) TIME=0; 
for when in $(echo %WHENS:0%); do
  if [[ $((NOW - when)) -le 1 ]]; then break; fi
  TIME=$when
done
for when in $(echo %WHENS:0%); do
  if   [[ $when -le $NOW ]] then  xmeter time $when; sleep %SLEEP:5%; fi
  done;; 
time) TIME=%TIME:0%

main=/sapp/run/ext
case $TIME in 
1350) path=$main/DC1/00;;
0120) path=$main/DC1/12;;
0205) path=$main/DC1/18;;
1405) path=$main/DC1/06;;
1605) path=$main/DA1/12;;
2205) path=$main/DA1/18;;
0405) path=$main/DA1/00;;
1005) path=$main/DA1/06;;
*) xabort;;
esac

case %ECF_PORT:0% in
0) login="set SMS_PROG %SMS_PROG:0%; login %SMSNODE:0% $USER 1"
   cdp -c "$login; force -r complete $path";;
*) ecflow_client --force=complete recursive $path;;
esac

;;
esac
%include <endt.h>
#include <step2.h>
"""

    create_wrapper("sapp.sms", content)

    content = """
%manual 
manual - this task is automatically created by cray.py

expected to run in real-time mode only
%end
#include <qsub.h>
#include <step1.h>
%include <trap.h>

if [[ $USER == map ]]; then
cd /tmp/map/work/p4/metapps/suites/o/def
make etest | grep -v Node::addVariable:
make teste | grep -v Node::addVariable:

TEST="" make oe edae lawe lbce rd mce mofce
TEST= "" make e eeda elbc elaw 

fi

cd /home/ma/emos/def/o/def
# make etest| grep -v Node::addVariable:
# make teste| grep -v Node::addVariable:

TEST="" make oe edae lawe lbce rd mce mofce| grep -v Node::addVariable:
TEST="" make e eeda elbc elaw | grep -v Node::addVariable:

make doc
make covtest
make cov| grep -v Node::addVariable:

# e o rd

%include <endt.h>
#include <step2.h>
"""
    create_wrapper("maker.sms", content)

    content = """#!/bin/ksh
%manual 
manual - this task is automatically created by cray.py
expected to run in real-time mode only
%end
#include <qsub.h>
#include <step1.h>
%include <trap.h>

KIND=%KIND:when%
NOW=$(date +%%H%%M)

YMD=%YMD:20010101%
BASETIME=%YMD:20010101%%EMOS_BASE:00%
BASE=%EMOS_BASE:00%
STREAM=%STREAM:elda%
MEMBER_FROM=%MEMBER_FROM:%
MEMBER=%MEMBER:%
EXPVER_FROM=%EXPVER_FROM:9093%
EXPVER=%EXPVER:9963%

case $HOST in
c2*) FDB_ROOT=%STHOST:/s2o1%/ma_fdb;;
cc*) FDB_ROOT=%STHOST:/sc1%/tcwork/emos/ma_fdb;;
*) exit 1;;
esac

cd $FDB_ROOT
mkdir fix
cd fix
scp emos@%SCHOST_BKUP:cca:$FDB_ROOT/:od:$STREAM:g:$EXPVER_FROM:$YMD::/:${EMOS_BASE}00:$MEMBER_FROM:*: .

cat > rules.txt <<EOF
set expver="$EXPVER";
write;
EOF

for f in $(ls :${EMOS_BASE}00:$MEMBER_FROM:*:); do
  grib_filter -o $f.0001 rules.txt $f
done

echo "mkdir -p test; for f in $(ls :${EMOS_BASE}00:$MEMBER_FROM:*:); do FDB_ROOT=$(pwd)/test grib2fdb $f; done"
echo "mv test/*$STREAM:*/* $FDB_ROOT/:od:$STREAM:g:$EXPVER:$YMD::/."
done

case $KIND in 
VERSION) echo;;
MEMBER) echo;;
*) xabort;;
esac

;;
esac
%include <endt.h>
#include <step2.h>
"""

    create_wrapper("safety.sms", content)

def create_de():
    try:    os.stat(wdir + "/de")
    except: os.makedirs(wdir + "/de")       
    task = open(wdir + "/de/setup.ecf", 'w')
    print('''#!/bin/ksh
%include <head.h>
%ecfmicro ~
~includenopp <scripts/setup.run>
~ecfmicro %
%include <tail.h>
''', file=task)

    task = open(wdir + "/de/model.ecf", 'w')
    print('''#!/bin/ksh
# QSUB -lh  %THREADS:1%
# QSUB -lPv %NPES:1%
# @ tasks_per_node = 64
%include <head.h>
%ecfmicro ~
~includenopp <scripts/model.run>
~ecfmicro %
%include <tail.h>
''', file=task)

    logdir = "/s2o2/emos_dir"
    return Family("de").add(
        Defcomplete(),
        Repeat("YMD", 20100101, 20121212),
        Variables(ECF_EXTN= ".ecf",
                  ACCOUNT= "oesu",
                  SCHOST= "cca",
                  ECF_LOGHOST= "cca",
                  ECF_OUT= logdir,
                  LOGDIR= logdir,
                  QUEUE= "ns",
                  TURTLES= 1000*'I LIKE ',
                  ),
        Task("setup"),
        Task("obs").add(Trigger("setup==complete")),
        Task("namelist").add(Trigger("obs==complete")),
        Task("model").add(
            Variables(NPES= 1,
                      THREADS= 1,
                      QUEUE= "np",),
            Trigger("namelist==complete")),
        Task("postproc").add(Trigger("model==complete")),
        )

def create_smhi():
    submit = open(wdir + "/smhi_run.sh", 'w')
    print('''#!/bin/ksh
set -eux
ECF_NAME=$1
ECF_PASS=$2
ECF_JOB=$3
ECF_JOBOUT=$4
ECF_HOST=$5
ECF_PORT=$6
export ECF_NAME ECF_PASS ECF_HOST ECF_PORT ECF_JOB ECF_JOBOUT
echo "#MSG: START USER SCRIPT"
. $ECF_JOB # >> $ECF_JOBOUT 2>&1
''', file=submit)

    man = open(wdir + "/smhi_man.h", 'w')
    print('''
an example for include manual page
''', file=man)
    man.close()

    shell = open(wdir + "/smhi.sh", 'w')
    print('''#!/bin/ksh
set -eux
# use ecflow || 
export PATH=/usr/local/apps/ecflow/current/bin:$PATH
export SIGNAL_LIST='1 2 3 4 5 6 7 8 13 15 24 31'
ecflow_client --init=$$

ERROR() {
  ecflow_client --abort # raison:trap
  echo "#WAR: TRAPPING ERROR"
  trap 0  
  uname -a; date; times
  sleep 0
  exit 0
}
trap ERROR 0 $SIGNAL_LIST

ecflow_client --event=1
ecflow_client --meter=step 10

echo "simulates an error"; exit 1
echo "simulates an early exit: expect this to be trapped and reflected as error"
exit 0 # test bubbling up an error through ssh

''', file=shell)
    wrapper = open(wdir + "/smhi_v.sms", 'w')
    print('''
%manual
  example of minimalist script: only variables, wrapper is added at submission time
%end
#SET ECF_NAME=%ECF_NAME%
#SET ECF_HOST=%ECF_HOST%
#SET ECF_PORT=%ECF_PORT%
#SET ECF_PASS=%ECF_PASS%
#SET ECF_JOB=%ECF_JOB%
#SET ECF_JOBOUT=%ECF_JOBOUT%
#SET ECF_TRYNO=%ECF_TRYNO%
#SET SCRIPT_PATH=%SCRIPT_PATH%
#SET SCRIPT_NAME=%SCRIPT_NAME%
''', file=wrapper)

    submit = open(wdir + "/smhisubmit.sh", 'w')
    print('''#!/bin/ksh
set -eux
ECF_JOB=$1
grep "#SET " $ECF_JOB | sed -e "s:#SET ::" > $ECF_JOB.set
grep "#SET " $ECF_JOB > $ECF_JOB.var
. $ECF_JOB.set
echo "#!/bin/ksh" > $ECF_JOB
cat $ECF_JOB.set $ECF_JOB.var >> $ECF_JOB
cat >> $ECF_JOB <<\@@
set -eux
export ECF_NAME ECF_PASS ECF_HOST ECF_PORT
export PATH=/usr/local/apps/ecflow/current/bin:$PATH
export SIGNAL_LIST='1 2 3 4 5 6 7 8 13 15 24 31'
ERROR() {
  ecflow_client --abort trap
  echo "#WAR: TRAPPING ERROR"
  trap 0 
  uname -a; date; times
  exit 1
}
trap ERROR 0 $SIGNAL_LIST

if [[ ${USE_SSH_SUB:=0} == 1 ]]; then
ssh $SCHOST mkdir -p $(dirname $ECF_JOBOUT)
ssh $SCHOST -R$ECF_PORT:$ECF_HOST:$ECF_PORT ''' + wdir + '''/smhi_run.sh \
  $ECF_NAME $ECF_PASS $SCRIPT_PATH/$SCRIPT_NAME $ECF_JOBOUT $ECF_HOST $ECF_PORT
echo "#MSG: ssh exits with $?"
else
ecflow_client --init=$$

i=0
while [[ $i -lt ${STEPS:=1} ]]; do
  ecflow_client --meter=step $i
  ((i=i+1))
done

ecflow_client --event=1

fi

wait
trap 0
ecflow_client --complete
uname -a; date; times
exit 0
@@

chmod 755 $ECF_JOB
$ECF_JOB > $ECF_JOBOUT 2>&1 &

''', file=submit)

    os.system("chmod 755 " + wdir + "/smhisubmit.sh")
    wrapper = open(wdir + "/smhi.sms", 'w')
    os.system("chmod 755 " + wdir + "/smhi_run.sh")
    cmd = "ln -sf %s/test.sms " % wdir + wdir + "/"
    os.system(cmd + "a.sms")
    os.system(cmd + "a1.sms")
    os.system(cmd + "a11.sms")
    print("##cmd", cmd)

    print('''#!/bin/ksh
%manual 
%include <smhi_man.h>
%end
set -eux
# use ecflow ||
export PATH=/usr/local/apps/ecflow/current/bin:$PATH
export SIGNAL_LIST='1 2 3 4 5 6 7 8 13 15 24 31'
ECF_NAME=%ECF_NAME%
ECF_PASS=%ECF_PASS%
ECF_HOST=%ECF_HOST%
ECF_PORT=%ECF_PORT%
export ECF_NAME ECF_PASS ECF_HOST ECF_PORT

ERROR() {
  ecflow_client --abort # raise:trap
  echo "#WAR: TRAPPING ERROR"
  trap 0  
  uname -a; date; times; sleep 0
  exit 1
}

SLEEP=%SLEEP:0%
trap ERROR 0 $SIGNAL_LIST

if [[ %USE_SSH_SUB:0% == 1 ]]; then
ssh %SCHOST% mkdir -p $(dirname %ECF_JOBOUT%)
ssh %SCHOST% -R%ECF_PORT%:%ECF_HOST%:%ECF_PORT% ''' + wdir + '''/smhi_run.sh \
  %ECF_NAME% %ECF_PASS% %SCRIPT_PATH%/%SCRIPT_NAME% %ECF_JOBOUT% %ECF_HOST% %ECF_PORT%
  echo "#MSG: ssh exits with $?"
else
ecflow_client --init=$$

i=0
while [[ $i -lt ${STEPS:=1} ]]; do
  ecflow_client --meter=step $i
  ((i=i+1))
done

ecflow_client --event=1

fi

wait
trap 0
ecflow_client --complete
uname -a; date; times
exit 0
''', file=wrapper)

def dummy(): return Task("test")

def logsvr(kind): return Task("logsvr").add(
    Label("man", "kind may be start, stop, ping"),
    Label("info", ""),
    Variables(KIND= kind))

def limits():
    return (Family("Test").add(Defcomplete(),
                               Label("info", "test find reg exp, case sens"),
                               Task("Ecmwf"),
                               Task("ECMWF"),
                               Task("ecmwf"),
                               Task("ecm")),
            Family("limits").add(
            Defcomplete(),
            Limit("lim", 10),
            Limit("test", 10),
            Limit("hpc", 10),
            Limit("cca", 10),
            Limit("ccb", 10),
            # Limit("cct", 10),
            # Limit("c2a", 10),
            # Limit("c2b", 10),
            Inlimit(ic.psel() + "/limits:lim")))

def repeat():
    return Family("repeat").add(
        Defcomplete(),
        Family("int").add(
            Repeat("YMD", 21000101, 21001212, 1, "integer"),
            dummy()),
        Family("date").add(
            Repeat("YMD", 21000101, 21001212, 1, "date"),
            dummy()),
        Family("string").add(
            Repeat("YMD", ["1", "2", "3", "4", "5"], kind="string"),
            dummy()),
        Family("enum").add(
            Repeat("YMD", "1 2 3 4 5".split(), kind="enum"),
            dummy()),
        )
        

def fam_ui():
    return Family("ui").add(
        Defcomplete(),
        repeat(),
        
        Variables(VAR= 1),

        Family("limit").add(Limit("mutex", 1),
                            InLimit("limit:mutex"),
                            Defstatus("complete"),
                            dummy()),

        Family("autocancel").add(AutoCancel("+01:00"),
                                 dummy()),

        dummy().add(Meter("step", -1, 100),
                          Event(1),
                          Label("info", ""),
                          Late("-s 00:10 -c 01:00"),
                          ),

        Family("complete").add(Complete("test:1"),
                               Trigger("test==complete")),

        Task("time").add(Time("10:00"),
                         Today("11:00"),
                         Cron("12:00 13:00 00:05"),
                         ),
        Task("date").add(Date("01.*.*"),)
        )

def smhi():
    return Family("smhi").add(
        Defcomplete(),

        Variables(ECF_FILES= wdir,
                  QUEUE= "none",
                  LOGDIR= jdir, # "/vol/lxop_emos_nc/output",
                  MANPATH= "%ECF_SCRIPT%",
                  ECF_JOB_CMD= "%ECF_JOB% > %ECF_JOBOUT% 2>&1",
                  ECF_HOME= jdir, # "/tmp/smhi",
                  ECF_INCLUDE= wdir,),

        acq(),

        Family("ex1").add(Cron("00:00 23:59 00:03"),
                          Task("scan").add(Label("info", "one min"),
                                           Event("1")),
                          Task("proc").add(Label("info", "one min"),
                                           Trigger("scan:1"),
                                           Complete("scan eq complete and not scan:1"))),
        Family("ex2").add(
            Task("scan").add(Label("info", "one min"),
                             Cron("00:00 23:59 00:03"),
                             Event("1")),
            Family("consume").add(
                Label("info", "event is set by scan task"),
                Event("1"),
                Cron("00:00 23:59 00:05"),
                Trigger("consume:1"),                
                Task("proc1").add(Trigger("1==0", Complete("1==1"))),
                Task("proc2").add(Trigger("1==0", Complete("1==1"))),
                )),
        Family("name").add(
            Variables(ECF_JOB_CMD= SUBMIT,
                      ECF_KILL_CMD= KILL,
                      ECF_STATUS_CMD= STATUS,
                      ECF_CHECK_CMD= CHECK,
                      # ECF_EXTN= ".sms",
                      ACCOUNT=  "UNSET",                      
                      ECF_INCLUDE= idir,
                      ECF_HOME= jdir,
                      ECF_FILES= wdir,
                      TOPATH= udir + "/logs",
                      SLEEP=    1,
                      SCHOST= "ibis",
                      ECF_TRIES= 1,),
            Task("a").add(Meter("step", -1, 100), Event("1")),
            Task("a1").add(Meter("step", -1, 100), Event("1")),
            Task("a11").add(Meter("step", -1, 100), Event("1")),
            ),
        
        Family("ssh").add(
            Variables(USE_SSH_SUB= 1,
                      SCRIPT_PATH= wdir,
                      SCRIPT_NAME= "smhi.sh"),            
            smhi_unit("ibis", "ibis", "none", jdir),
            smhi_unit("pikachu", "pikachu", "none", jdir),
            ),

        Family("alter").add(
            Variables(USE_SSH_SUB= 1,
                      ECF_JOB_CMD= "/home/ma/map/course/cray/smhisubmit.sh %ECF_JOB%",
                      SCRIPT_PATH= wdir,
                      SCRIPT_NAME= "smhi.sh"),            
            smhi_unit_alter("ibis", "ibis", "none", jdir),
            smhi_unit_alter("pikachu", "pikachu", "none", jdir),
                          )
        )        
                          
def acq():
    name = "acq_ex"

    fam = Family(name).add(
        Defcomplete(),
        Family("times").add(
            dummy().add(
                Time("10:00 20:00 00:30"))),
        Complete(name + "/data eq complete"),
        Task("data").add(
            Event("ready"),
            Trigger("rt/wait:data"),
            ),
        Family("rt").add(
            Complete("data:ready"),
            Task("wait").add(
                Event("data"),
                # Cron("10:00 12:00 00:05")
                )),
        Task("late").add(
            Time("11:00"),
                Trigger("not data:ready")))
    return fam


HCRAY="cct"

SUBMIT=ic.SUBM + " %USER% %SCHOST% %ECF_JOB% %ECF_JOBOUT% submit"
GSUB=SUBMIT.replace("%SCHOST%", "%SCHOST% %ECF_RID%")
KILL=GSUB.replace(" submit", " kill")
STATUS=GSUB.replace(" submit", " status")
CHECK=STATUS


def unit(name, schost, queue, rdir, account="", leaf=True, add=None):
    return Family(name).add(
        If(account != "",
           Variables(ACCOUNT= account)),
        Variables(QUEUE=    queue, 
                  SCHOST=   schost,
                  ECF_OUT=  rdir,
                  LOGDIR=   rdir,),        
        If (leaf, Task("test").add(
                add,
            Event(1),
            Meter("step", -1, 120, 100),
            Label("info", "nop"))))

def smhi_nit(name, schost, queue, rdir, account=""):
    return Family(name).add(
        If(account != "", Variables(ACCOUNT= account)),
        Variables(QUEUE=    queue, 
                  SCHOST=   schost,
                  ECF_OUT=  rdir,
                  LOGDIR=   rdir,),        
        Task("smhi").add(
            Event(1),
            Meter("step", -1, 120, 100),
            Label("info", "nop")))

def smhi_unit(name, schost, queue, rdir, account=""): return None

def smhi_unit_alter(name, schost, queue, rdir, account=""):
    return Family(name).add(
        If(account != "", Variables(ACCOUNT= account)),
        Variables(QUEUE=    queue, 
                  SCHOST=   schost,
                  ECF_OUT=  rdir,
                  LOGDIR=   rdir,),        
        Task("smhi_v").add(
            Event(1),
            Meter("step", -1, 120, 100),
            Label("info", "nop")))

def turkey_main(day, path):
    def runs(num, time):
        return Family('run%d' % num).add(
               Trigger(path + '/deps/%s eq complete' % time),
            
               Family('feed_meb').add(
                Task('strip_message'),                  

                Task('insert_meb').add(
                    Trigger('strip_message eq complete'),
                  ),
               ),               
               Family('feed_reb').add(
                  Task('decode_synop').add(
                     Trigger('../feed_meb/insert_meb eq complete'),
                  ),
                  Task('decode_temp').add(
                     Trigger('../feed_reb/decode_synop eq complete'),
                  ),
                  Task('decode_ship').add(
                     Trigger('../feed_reb/decode_temp eq complete'),
                  ),
                  Task('decode_metar').add(
                     Trigger('../feed_reb/decode_ship eq complete'),
                  ),
               ),               
               Family('rdb').add(
                  Task('rdb_synop').add(
                     Trigger('../feed_reb/decode_synop eq complete'),
                  ),
                  Task('rdb_temp').add(
                     Trigger('../feed_reb/decode_temp eq complete'),
                  ),
                  Task('rdb_ship').add(
                     Trigger('../feed_reb/decode_ship eq complete'),
                  ),
                  Task('rdb_metar').add(
                     Trigger('../feed_reb/decode_metar eq complete'),
                  ),
               ),               
            ),            

    def task_trigger(name, condition):
        return Task(name).add(Trigger(condition))

    def echos():
        echo = { 
            "cleandb" : "0131",
            "cleanerrdb": "0136",
            "silmeb": "0141",
            "silreb": "0146",
            "run1": "0221",
            "run2": "0421",
            "run3":"0621",
            "run4": '0821',
            'run5': '1021',
            'run6': '1321',
            'run7': '1421',
            'run8': '1621',
            'run9': '1821',
            'run10': '2121',
            'eksikveritamamla': '2201',
            'uydutsm': '0201',
            'model_all_gunluk_tsm': '0131',
            'radar_lokasyon_tip_tsm': '0501',
            'retObsAndFtp_for_00': '0231',
            'retObsAndFtp_for_18': '0241',
            'retObsAndFtp_for_12': '1541',
            'retObsAndFtp_for_06': '1101',
            'retObsAndFtp_for_06gungeri': '0801',
            'uydureport': '1655',
            'radarreport': '0701',
            'backuplog': '2301',
            'cleanup': '2316'
            }
        out = []

        for key in sorted(echo, key=echo.get, reverse=False): # order by value
            # for key in sorted(echo.keys()): # order by key
            out.append(task_trigger("echo_%s" % key, 
                                    path + "/deps/%s" % echo[key] + " eq complete"))
        return out
        
    return Family('main').add(
            Trigger('deps/0000 eq complete'),
         
            Family('clean').add(
               task_trigger('cleandb',
                            path + '/deps/0130 eq complete'),

               Task('clean_errdb').add(
                  Trigger(path + '/deps/0135 eq complete'),
               ),
               Task('silmeb').add(
                  Trigger(path + '/deps/0140 eq complete'),
               ),
               Task('silreb').add(
                  Trigger(path + '/deps/0145 eq complete'),
               ),
            ),         
            runs(1, "0220"),
            runs(2, "0420"),
            runs(3, "0620"),
            runs(4, "0820"),
            runs(5, "1020"),
            runs(6, "1320"),
            runs(7, "1420"),
            runs(8, "1620"),
            runs(9, "1820"),
            runs(10, "2120"),
            Family('model').add(
               Task('eksikveritamamla').add(
                  Trigger(path + '/deps/2200 eq complete'),
               ),
            ),            
            Family('tsmarc').add(
               Task('uydu_tsm').add(
                  Trigger(path + '/deps/0200 eq complete'),
               ),
               Task('model_all_gunluk_tsm').add(
                  Trigger(path + '/deps/0130 eq complete'),
               ),
               Task('radar_lokasyon_tip_tsm').add(
                  Trigger(path + '/deps/0500 eq complete'),
               ),
            ),            
            Family('ftpsenddata').add(
               Task('retObsAndFtp_for_00').add(
                  Trigger(path + '/deps/0230 eq complete'),
               ),
               Task('retObsAndFtp_for_18').add(
                  Trigger(path + '/deps/0240 eq complete'),
               ),
               Task('retObsAndFtp_for_12').add(
                  Trigger(path + '/deps/1540 eq complete'),
               ),
               Task('retObsAndFtp_for_06').add(
                  Trigger(path + '/deps/1100 eq complete'),
               ),
               Task('retObsAndFtp_for_06gungeri').add(
                  Trigger(path + '/deps/0800 eq complete'),
               ),
               Task('uydu_report').add(
                  Trigger(path + '/deps/1654 eq complete'),
               ),
               Task('radar_report').add(
                  Trigger(path + '/deps/0700 eq complete'),
               ),
            ),            
            Family('cleanup').add(
               Task('backuplog').add(
                  Trigger(path + '/deps/2300 eq complete'),
               ),
               Task('cleanup').add(
                  Trigger(path + '/deps/2315 eq complete'),
               ),
            ),            
            Family('echo').add(
               Variables(
                  ECF_FILES= '/pp1/bin/dpp_sms/echo',),

               echos(),           
         ))


def turkey_deps(num):
    def item(hhmm):
        return  Family(hhmm).add(
               Complete("./%s/compl eq complete and " % hhmm + 
                        './%s/dummy eq queued' % hhmm),
            
               Task('dummy').add(
                  Cron('-w %d' % num + ' ' + hhmm[:2] + ":" + hhmm[-2:]),
               ),
               Task('compl').add(
                  Complete("dummy eq active or " +
                           "dummy eq submitted or " +
                           'dummy eq complete'),
                  Trigger('1 eq 0'),
               ),
            )

    items = [ "0000",
"0130",
"0135",
"0140",
"0145",
"0220",
"0420",
"0620",
"0820",
"1020",
"1320",
"1420",
"1620",
"1820",
"2120",
"2200",
"0200",
"0500",
"0230",
"0240",
"1540",
"1100",
"0800",
"1654",
"0700",
"2300",
"2315",
"0131",
"0136",
"0141",
"0146",
"0221",
"0421",
"0621",
"0821",
"1021",
"1321",
"1421",
"1621",
"1821",
"2121",
"2201",
"0201",
"0501",
"0231",
"0241",
"1541",
"1101",
"0801",
"1655",
"0701",
"2301",
"2316", ]
    def loop_items(): 
        out = []
        for hhmm in items: 
            out.append(item(hhmm))
        return out 

    return Family('deps').add(
            Variables(
               ECF_FILES= '/pp1/bin/dpp_sms/haftalik',),
            loop_items())

def turkey_loop(suite_name):
    out = []
    days = [ "Pazartesi", "Sali", "Carsamba", "Persembe", "Cuma", "Cumartesi", "Pazar"]
    days = [ "mon", "tue", "wed", "thu", "fri", "sat", "sun"]
    start = 20120723; end=20301231 # beware addition below does not consider start as a date
    path = "/%s/weekly/" % suite_name

    for num in range(7): 
        out.append(Family(days[num]).add(
                Repeat(kind='date', name='YMD', 
                       start= start + num, end=end, step=7),
                turkey_deps(1),
                turkey_main(days[num], path=path + days[num]),))
    return out

def turkey(suite_name):
    return    Family('weekly').add(
        Label("repeat", "run one family per week day"),
        Label("time_trigger", "replace time dependencies with trigger"),
        Label("aborted", "one aborted task requests being fixed within one week"),
        Defcomplete(),
        turkey_loop(suite_name),)

##########################################################################
## PURE

class Languages(object):
    def __init__(self): pass 
    def task(self): return None
    def scripts(self): pass

class Python(Languages):
    def task(self):
        return (
            Task("pure_python").add(
                Variables(ECF_MICRO= "$",
                          ECF_INCLUDE= wdir,
                          SCHOST= "localhost",
                          ECF_JOB_CMD= "use ecflow; $ECF_JOB$ > $ECF_JOBOUT$ 2>&1 &"),
        Event("1"),
        Label("info", ""),
        Meter("step", -1, 100),
        # Defstatus("complete"),
        ),

        Task("python").add(
        Variables(ECF_MICRO= "$",
                  ECF_INCLUDE= wdir,
                  SCHOST= "localhost",
                  ECF_JOB_CMD= "$ECF_JOB$ > $ECF_JOBOUT$ 2>&1 &"),
        Event("1"),
        Label("info", ""),
        Meter("step", -1, 100),
        # Defstatus("complete"),
        ), )

class Perl(object):
    def task(self):
        return Task("perl").add(
            Label("todo", "time-stamp PS4, set eux, exit 0 1,"),
            Variables(ECF_MICRO= "^",
                      ECF_INCLUDE= wdir,
                      SCHOST= "localhost",
                      ECF_JOB_CMD= "^ECF_JOB^ > ^ECF_JOBOUT^ 2>&1 &"),
            Event("1"),
            Label("info", ""),
            Meter("step", -1, 100),
            # Defstatus("complete"),
            )

class Ruby(object):
    def task(self):
        return Task("ruby").add(
            Label("todo", "time-stamp PS4, set eux, exit 0 1,"),
            Variables(ECF_MICRO= "^",
                      ECF_INCLUDE= wdir,
                      SCHOST= "localhost",
                      ECF_JOB_CMD= "^ECF_JOB^ > ^ECF_JOBOUT^ 2>&1 &"),
            Event("1"),
            Label("info", ""),
            Meter("step", -1, 100),
            Defstatus("complete"),
            )

    def scripts(self):
        head = """#!/usr/bin/env ruby"""
        tail = """ """
        body = """ """
        create_wrapper("head.rb", head)
        create_wrapper("tail.rb", tail)
        create_wrapper("ruby.sms", body)

ruby = Ruby()
perl = Perl()
pyth = Python()

WDIR= "/scratch/ma/map/sms"
HLOG= "/home/ma/map/logs"

hosts = {"lxb" : { "COMPILER": "gcc-4.3", # linux64
                   "SCHOST": "lxb", 
                   "QUEUE_EPILOG": "serial",
                   "LOGDIR": HLOG, "ECF_OUT": HLOG, 
                   "QUEUE": "serial",
                   "WDIR": WDIR, }, 
         "ibis" : { "COMPILER": "gcc",
                    "SCHOST": "ibis", 
                    "WDIR": WDIR, }, 
         "lxop" : {  "COMPILER": "gcc-4.3",
                    "SCHOST": "lxop",
                     "QUEUE": "test", 
                     "QUEUE_EPILOG": "test",
                     "LOGDIR": HLOG, "ECF_OUT": HLOG, 
                     "COMPILE_DIR": "/gpfs/lxop/emos_data/sms", 
                    "WDIR": "/gpfs/lxop/emos_data/sms", 
                     }, 
         "opensuse103": { "COMPILER": "gcc-4.5",
                         "SCHOST": "opensuse103", 
                     "LOGDIR": HLOG, "ECF_OUT": HLOG, 
                         "WDIR": WDIR, }, 
         "opensuse113": { "COMPILER": "gcc-4.5",
                         "SCHOST": "opensuse113", 
                     "LOGDIR": HLOG, "ECF_OUT": HLOG, 
                         "WDIR": WDIR, }, 
         "opensuse131": { "COMPILER": "gcc-4.8",
                         "SCHOST": "opensuse131", 
                     "LOGDIR": HLOG, "ECF_OUT": HLOG, 
                         "WDIR": WDIR, }, 
         "ecgb": { "COMPILER": "gcc-4.4.7", # redhat
                   "QUEUE": "normal", 
                   "QUEUE_EPILOG": "normal",
                   "SCHOST": "ecgb", 
                     "LOGDIR": HLOG, "ECF_OUT": HLOG, 
                    "WDIR": WDIR, }, 
         "lxc": { "COMPILER": "gcc-4.4.7", # redhat
                   "QUEUE": "normal", 
                   "QUEUE_EPILOG": "normal",
                   "SCHOST": "lxc", 
                   "LOGDIR": HLOG, "ECF_OUT": HLOG, 
                   "WDIR": WDIR, }, 
         "cct": { "COMPILER": "gcc",
                  "QUEUE_EPILOG": "ns",
                  "SCHOST": "cct", 
                  "LOGDIR": HLOG, "ECF_OUT": HLOG, 
                  "COMPILE_DIR": "/home/ma/map/sms", 
                  "WDIR": "/home/ma/map/sms", }, 
         "cca": { "COMPILER": "gcc",
                  "QUEUE_EPILOG": "ns",
                      "SCHOST": "cca", 
                       "LOGDIR": HLOG, "ECF_OUT": HLOG, 
                       "WDIR": "/home/ma/map/sms", }, 
         "ccb": { "COMPILER": "gcc",
                  "QUEUE_EPILOG": "ns",
                      "SCHOST": "ccb", 
                       "LOGDIR": HLOG, "ECF_OUT": HLOG, 
                      "WDIR": "/home/ma/map/sms", 
                  "COMPILE_DIR": "/home/ma/map/sms", 
                  },          }

class Compile(object):

    def __init__(self):
        pass

    def compile(self, host=None, add=None, but=None):
        version = "4.4.15"
        if host is None: dest = list(hosts.keys())
        else: dest = (host, )
        out = []
        for host in sorted(dest):
            if not but: pass
            elif host == but or host in but: continue
            out.append(Family(host).add(
                    add,
                    Variables(hosts[host]),
                    Variables(ECF_FILES= wdir,
                              VERSION= version,
                              ECF_INCLUDE= idir,         ),
                    Task("get"),
                    Task("make").add(Trigger(["get"])),
                    Task("install").add(Defcomplete(),
                                     Trigger(["make"])),
                    Task("testit").add(Event("cdp"),
                                       Event("sms"),
                                       Event("child"),
                                       Event("xcdp"),
                                       Trigger(["make"])),
                    Task("child").add(Defcomplete(),
                                      Variables(SMS_VERSION= "4.4.15",),
                                      Label("info", ""),
                                      Event("1"),
                                      Meter("step", -1, 100)),
                    ))
        return out

    def main(self):
        self.create_tasks()
        return Family("sms").add(Family("compile").add(
            Defstatus("suspended"),
            # onws("localhost"),
            Variables(SCHOST= "localhost"),
            Task("prepare"),
            self.compile("ibis", add=Trigger("prepare==complete")),
            Family("remote").add(Trigger("prepare==complete"),
                                 self.compile(but="ibis"),)))

    def wrap(self, content):
        return '''#!/bin/ksh
# generated by cray.py
%%manual
  start - check - stop logsvr
%%end
%%include <qsub.h>
%%include <step1.h>
%%include <trap.h>
%s 
%%include <endt.h>
%%include <step2.h>''' % content

    def create_tasks(self):
        source = "%SOURCE_DIR:/tmp/map/work/git/sms%"
        compil = "%COMPILE_DIR:/scratch/ma/map/sms/$ARCH/%/%SCHOST:$HOST%"
        task = self.wrap("cd %s/sms; make clean; cd ../; tar -czf sms.tgz sms; " % source)
        create_wrapper("prepare.sms", task)
        
        task = self.wrap("comp=%s; " % compil + 
                         "mkdir -p $comp; cd $comp; scp ibis:%s/sms.tgz .;" % source +
'''rm -rf sms.tar sms || :
gzip -d sms.tgz; tar -xf sms.tar''')
        create_wrapper("get.sms", task)

        task = self.wrap("comp=%s;" % compil + 
                         '''cd $comp/sms; 
if [[ $ARCH == cray ]]; then
module switch PrgEnv-cray PrgEnv-intel
export CRAYPE_LINK_TYPE=dynamic
fi
make %TARGET:linux%''')
        create_wrapper("make.sms", task)

        task = self.wrap("comp=%s;" % compil +  
                         '''cd $comp/sms; make tester; 
mkdir -p %SMSHOME:/tmp/$USER%
SMS_PROG=900001 sms &
sleep 10
xevent sms
SMS_PROG=900001 smsping localhost || :
./cdp/cdp -c 'set SMS_PROG 900001; login $HOST $USER 1; terminate -y; exit 0'
xevent cdp
DISPLAY=ibis:0.0 ./xcdp/xcdp &
xevent xcdp
ln -sf ./cdp/child smsevent
./smsevent child
rm -f smsevent''')
        create_wrapper("testit.sms", task)

        task = self.wrap("xevent 1; xlabel info test; xmeter step 0; xmeter step 100;")
        create_wrapper("child.sms", task)

##########################################################################
## BARBER

def call_barber_passerby():
    # family passby
    # limit to one at a time entering the
    # shop to check if there is room to
    # stay in there
    return Task("passby").add(
        Cron("00:00 23:59 00:01"),
    # this one is regular
        Variables(ID= 0),
        Inlimit("limit:passby"))
    # other frequent passers by can be added here
# called by client generator task

def call_barber_client(id):
    return Family(id).add(
    # to be called by passby to create a new client
        Task("cut").add(
            Inlimit("../limit:barbers")),
    # there may be many barbers in the shop
    # but one cashier only
        Task("pay").add(
            Trigger("./cut == complete"),
            Inlimit("../limit:barbers"),
            Inlimit("../limit:cashiers")),
        Task("leave").add(
            Trigger("cut==complete and pay==complete"),
            If(id % 2 == 0,
               Defcomplete(),
               Autocancel( "0" ))))

# main definition
def call_barber_shop():
    return Family("shop").add(
        Defstatus("suspended"),
        Variables(NB_CHAIRS= 4),

        Family("limit").add(
            Defcomplete(),
            Limit("passby", 1),
            Limit("barbers", 1),
            Limit("cashiers", 1)),

        call_barber_passerby(),)

##########################################################################
## dinner
total_number = 5
life_expectency = 100
def requeued():
    return Repeat("iter", 1, life_expectency, kind= "integer")
    # possibility to requeue the task without duration limit
    # provided 02:00 is below the task duration:
    # cron 00:00 23:59 02:00

def triggered(ref, tor=None, tand=None):
    trg = ""
    tr2 = ""
    if tand: trg = " and %s" % tand
    if tor:  tr2 = " or %s" % tor
    return Trigger("(%s != active and %s != submitted and %s != aborted" 
                % (ref, ref, ref) + trg + ")" + tr2)

def philosopher(id, s1, s2, version):
    return Family("%d" % id).add(
        Task("eat").add(
            Inlimit("../limit:"+str(s1)),
            Inlimit("../limit:"+str(s2)),
            Event("release")),
        If( version == "normal",
                # eat task a3ttribute
                # TBD triggered("think")
                ( requeued(),

                  Task("think").add(
                        requeued(),
                        triggered( "eat", tor="eat:release or eat==complete")))),
        If( version == "simple",
                (Task("release").add(
                        Trigger("eat:release or eat==complete"),),
                 requeued())))

def call_dinner():
    out = Family("dinner").add(
        Variables(SLEEP= 10),
        Defstatus("suspended"),
        )
    for version in "simple normal".split():
        # Xcdp would fold parent family when limits change
        # we prefer to store them in a dedicated family then
        lim = Family("limit").add(
                    Defcomplete(),)
# for num 1 $total_number do num 1 $total_number do
# $total_number do num 1 $total_number do
        num = 1
        while num <= total_number:
            lim.add(Limit(str(num), 1))
            num += 1; # endfor
        ver = Family(version).add(lim)
        out.add(ver)
# for num 1 $total_number do num 1 $total_number do
# $total_number do num 1 $total_number do
        num = 1
        while num <= total_number:
            stick1 = num
            stick2 = 1 + ((num) % total_number)
            ver.add(philosopher( num, stick1, stick2, version))
            num += 1; # endfor
    return out

##########################################################################
## CONSUMER
# /home/ma/map/course/course2011/src/consumer.py
beg =  0
fin = 48
by  =  3
    
def not_consumer(): return Variables(CONSUME= "no")

def not_producer(): return Variables(PRODUCE= "no")

def ev():return (Event("p"), Event("c"))

def call_task(name, start, stop, inc):
    meter = None
    if start != stop:
        meter = Meter( "step", -1, int(stop))

    return Task(name).add(
        ev(),
        Variables(BEG= start,
                  FIN= stop,
                  BY= inc),
        meter)

def call_consumer(SELECTION):
    # defstatus suspended # avoid early start
    def consume1(leap = 1, leap_nb = 3):
        out = []
        while leap <= leap_nb:
            leap_beg = beg + by * (leap - 1)
            leap_by = by * leap_nb
            out.append(Family("%d" % leap).add(
                    call_task( "consume", "'%STEP%'", "'%STEP%'", leap_by).add(
                        Repeat("STEP", leap_beg, fin, leap_by, kind="integer"),
                        # same trigger as consume0
                        # Trigger("(%s and (consume:STEP le %s1/produce:STEP)) or " % (lead, prod) + 
                        #         "(not %s and (consume:STEP le %s0/produce:step))" % (lead, prod))
                        Trigger("(consume:STEP le %s1/produce:STEP)" % prod),
                       )))
            leap += 1;
        return out

    def consume2(idx, fin):
        out = []
        while idx <= fin:  
            out.append(Family("%03d" % idx).add(
                    call_task( "consume", idx, idx, by).add(
                        Variables(STEP= idx),
                        # Same trigger as consume0
                        # Trigger("(%s and (consume:STEP le %s1/produce:STEP)) or " % (lead, prod) +
                        #         "(not %s and (consume:STEP le %s0/produce:step))" % (lead, prod))))
                        Trigger("consume:STEP le %s1/produce:STEP" % prod),
                       )))
            idx += by
        return out

    lead = "/%s/consumer/admin/leader:1" % SELECTION
    prod = "/%s/consumer/produce" % SELECTION
    if 1: return Family("consumer").add(Defcomplete()) # FIXME
    return Family("consumer").add(
        Defcomplete(),

        Variables(SLEEP= 10,
                  PRODUCE= "no",
                  CONSUME= "no"),

             Family("limit").add(
                Defcomplete(),
                Limit("consume", 7),),

             Family("admin").add(
                # set manually with Xcdp or alter the event 1 so
                # that producer 1 becomes leader
                # default is producer0 leads
                Task("leader").add(
                    Event("1"),
                    Defcomplete())),
             # text this task is dummy task not designed to run

             # default : task does both :
                 Variables(PRODUCE= "yes",
                           CONSUME= "yes"),
             # this task will do both, ie serial

             call_task( "produce", beg, fin, by).add(
            Label("info", "do both at once"),),
             # this will loop inside the task, reporting
             # its processed step as a meter indication

             Family("produce0").add(
                not_consumer(),
                call_task( "produce", beg, fin, by )),
             # here, choice is to push a new task for each step

             Family("produce1").add(
                not_consumer(),
                call_task( "produce", '%STEP%', "%STEP%", by ).add(
                Repeat("STEP", beg, fin, by , kind="integer"), )),
             # PRB edit FIN '$((%STEP% + %BY%))'

             Family("consume").add(
                not_producer(),
                Inlimit("limit:consume"),
                Variables(CALL_WAITER= 1,
                          # $step will be interpreted in the job!
                          TRIGGER= "../produce:step -gt consume:$step or ../produce eq complete"),
                call_task( "consume", beg, fin, by ).add(
                )),
             
             Family("consume0or1").add(
                not_producer(),
                Inlimit("limit:consume"),
                call_task( "consume", "%STEP%", "%STEP%", by, ),
                Repeat( "STEP", beg, fin, by, kind="integer"),
                Trigger("(%s and (consume0or1:STEP le %s1/produce:STEP)) or "
                        % (lead, prod) +
                        "(not %s and (consume0or1:STEP le %s0/produce:step))"
                        % (lead, prod))),

             Family("consume1").add(
                not_producer(),
                Inlimit("limit:consume"),
                consume1()),

             Family("consume2").add(
    # loop is "exploded"
    # consume limit may be changed manually with Xcdp
    # to reduce or increase the load
                Inlimit("limit:consume"),
                not_producer(),
                consume2(beg, fin)))

##########################################################################
class Cray(ic.SeedOD):
    def __init__(self):
        super(Cray, self).__init__()

    def setup(self, node):
        account= "UNSET"
        global user, host
        if user == "emos" or host in ("ode", "eode", "ecf"):
            rdir = "/vol/lxop_emos_nc/output"
            user= "emos"
            account= "oesu"
        else: pass
        self.defs.add_extern("/o/main:YMD")

        node.add(
            Clock("real"),
            Defstatus("suspended"),
            Label("info", "test,logsvr tasks generated by cray.py"),
            Event("1"),
            Meter("step", -1, 100),
            Repeat(kind="day", step=1),
            limits(),

            Variables(ECF_JOB_CMD= SUBMIT,
                      ECF_KILL_CMD= KILL,
                      ECF_STATUS_CMD= STATUS,
                      ECF_CHECK_CMD= CHECK,
                      # ECF_EXTN= ".sms",
                      ACCOUNT=  "UNSET",
                      SCHOST= "localhost",
                      ECF_INCLUDE= idir,
                      ECF_HOME= jdir,
                      ECF_FILES= wdir,
                      QUEUE= "ns", USER= user, LOGDIR= "/tmp",
                      TOPATH= udir + "/logs",
                      SLEEP=    1,     # x120
                      ECF_TRIES= 1,),

            Family("looper").add(Time("09:00"),
                                 Variables(SCHOST= "ibis"),
                                 dummy()), # .add(Defcomplete())),

            create_de(),

            smhi(),

            Compile().main(),

            Task("dummy").add(Time("16:00"),
                              Variables(SCHOST= "localhost",
                                        FAMILY= "unset",
                                        FAMILY1= "unset"),
                              Time("17:00"),),

            fam_ui(),

            # turkey(ic.selection()),

            sapp(),

            Family("maker").add(maker(),),

            call_consumer(ic.selection()).add(
                Label("repeat", "repeat may be attached to a family, or a task, to facilitate requeue or not"),
                Label("event", "event may be set by the task itself, by a user, by a third party job"),
                Label("trigger", "can be on the task-family node, or inside a task (wait)"),
                Label("deadlock", "produce0/produce complete with step=-1"),
                Label("latency", "between task wrapper creation, and new file visible from the server"),
                Label("ZZZ", "icon when the job is waiting the trigger condition"),
                Variables(QUEUE= "ns", USER= user, LOGDIR= "/tmp",
                          SCHOST="localhost"),),

            call_barber_shop().add(Defstatus("complete"),),

            call_dinner().add(Defstatus("complete"),),

            pyth.task(),

            perl.task(),

            ruby.task(),

            Family("blacklist").add(
                Event("cca"),
                Event("ccb"),
                Event("hps"),
                
                Defstatus("complete"))            )

        msg = "have you? setup ssh login, created remote directory" 
        msg += ", set START_LOGSVR=1 (edit task), once, to launch the logserver"
        if "eod" in host: user = "emos"
        print("#MSG: host, user", host, user)

        node = node.family("submit").add(
            unit("jonas", "ibis", "no", "/tmp/$USER/jobs", add=(Today("07:00"),
                                                                Today("21:00"))),
            Variables(USER=     user,
                      ACCOUNT=  account), # 

            # Time("09:00"),

            self.arg4(),
            self.burst(),
            # self.arg3(),
            # self.prb(),

            )

    def burst(self):
        sub = "/home/ma/emos/bin/smssubmit "
        out = []
        for num in range(0, 10):
            out.append(Family("%03d" % num).add(Task("test")))
        
        return Family("arg3").add(
            Defstatus("complete"),
            Variables(SCHOST= "cca-b"),
            out
            )

    def arg3(self):
        sub = "/home/ma/emos/bin/smssubmit "
        return Family("arg3").add(
            Defstatus("complete"),

            self.submits(),

            Variables(ECF_JOB_CMD= sub +"%USER% %SCHOST% %ECF_JOB%",
                      ECF_KILL_CMD= sub + "%USER% %SCHOST% %ECF_JOB% kill",
                      ECF_STATUS_CMD= sub + "%USER% %SCHOST% %ECF_JOB% status",
                      ),
            )

    def arg4(self):
        return Family("arg4").add(self.submits())

    def prb(self):
        sub = "/home/ma/emos/bin/smssubmit "
        return Family("prb").add(
            Defstatus("complete"),
            self.submits(),
            Label("info", "simulate prb at submission with % remaining"),
            Variables(ECF_JOB_CMD= sub +"%USER% %SCHOST% %ECF_JOB% %STTHOST%%ECF_JOBOUT%",
                      ECF_KILL_CMD= sub + "%USER% %SCHOST% %ECF_JOB% %STTHOST%%ECF_JOBOUT% kill",
                      ECF_STATUS_CMD= sub + "%USER% %SCHOST% %ECF_JOB%  %STTHOST%%ECF_JOBOUT% status",
                      ),                  )

    def submits(self):
        global host, user
        def parallel(pes=1, thr=1, hyp="no"):
            return Family("p%d_%d_%s" %(pes, thr, hyp)).add(
                Variables(NPES=pes, THREADS=thr,                           
                          HYPER=hyp), 
                Task("test").add(Label("info", ""), 
                                 Meter("step", -1, 120),
                                 Event(1)))
        ct1logs= "/sc1/tcwork/emos/logs"
        # ct1logs = '/home/ma/%s/logs' % user
        if ecflow: base = 1000
        else: base = 900000
        base = 1000
        if "eod" in host: user = "emos"

        out = [

            unit("cray", HCRAY, "ns", ct1logs, leaf=False).add(
                unit("ccp", "cca-login1", "ns", ct1logs).add(
                    Complete(ic.psel() + "/blacklist:cca"),),
                unit("cca", "cca-login1", "ns",    ct1logs).add(
                    logsvr("ping"),
                    Complete(ic.psel() + "/blacklist:cca"),),
                unit("ccb", "ccb-login1", "ns",    ct1logs).add(
                    logsvr("ping"),                    ),
                Family("parallel_a").add(
                    # Defcomplete(), ###
                    Complete(ic.psel() + "/blacklist:cca"),
                    Variables(SCHOST= "cca",
                              QUEUE= "os",
                              LOGHOST= "cca-il2",
                              ECF_OUT= ct1logs,
                              LOGDIR=  ct1logs,),
                    parallel(1, 1),
                    parallel(1, 24),
                    parallel(2, 12),
                    parallel(2, 24),
                    parallel(2, 15),
                    
                    parallel(1, 2, "yes"),
                    parallel(2, 24, "yes"),
                    parallel(24, 2, "yes"),
                    parallel(10, 5, "yes"),),

                Family("parallel_b").add(
                    Complete(ic.psel() + "/blacklist:ccb"),
                    Variables(SCHOST= "ccb",
                              QUEUE= "os",
                              LOGHOST= "ccb-il2",
                              ECF_OUT= ct1logs,
                              LOGDIR=  ct1logs,),
                    parallel(1, 1),
                    parallel(1, 24),
                    parallel(2, 12),
                    parallel(2, 24),
                    parallel(2, 15),
                    
                    parallel(1, 2, "yes"),
                    parallel(2, 24, "yes"),
                    parallel(24, 2, "yes"),
                    parallel(10, 5, "yes"),),

                If(user == "emos",
                   Variables(LOGHOST= HCRAY,
                             LOGPORT= base + get_uid()),
                   Variables(LOGHOST= HCRAY,
                             LOGPORT= base + get_uid())),
                )
            ]      

        lxop_out = "/vol/lxop_emos_nc/output"
        if user == "emos" or host in ("ode", "eode", "ecf"):
            ic.SUBM = "/home/ma/emos/bin/trimurtu"
            sms2ecf.subm = "/home/ma/emos/bin/trimurti"

            out += (
                Variables(LOGPORT= 9316),
                unit("cca", "cca", "ns", "/sc1/tcwork/emos_dir", "oesu").add(
                    Variables(STHOST= "/c1",
                              LOGHOST= "cca"),
                    logsvr("ping"),
                    Complete(ic.psel() + "/blacklist:cca"),),

                unit("ecgb", "ecgb", "emos", "/vol/emos_nc/output"),
                unit("lxc", "lxc", "emos", "/vol/emos_nc/output"),
                unit("lxab", "lxab", "serial", "/vol/emos_nc/output"),
                unit("lxop", "lxop", "emos", lxop_out
                     ).add(Variables(ECF_OUT= lxop_out,
                                     ECF_LOGHOST= "lxop",
                                     ECF_LOGPORT= 9316,
                                     LOGDIR=  lxop_out, )),
                unit("sappa", "sappa", "serial", "/vol/emos_nc/output"),
                unit("sappb", "sappb", "serial", "/vol/emos_nc/output"),
                unit("opensuse131", "opensuse131", "serial", "/vol/emos_nc/output"),
                Family("hps").add(
                    Complete(ic.psel() + "/blacklist:hps"),
                    unit("acq", "acq", "emos", "/tmp/emos"),
                    unit("pp1", "pp1", "emos", "/tmp/emos"),
                    unit("pp2", "pp2", "emos", "/tmp/emos"),
                    unit("pp3", "pp3", "emos", "/tmp/emos"),
                    unit("pp4", "pp4", "emos", "/tmp/emos"),
                    unit("bilbo", "bilbo", "emos", "/tmp/emos"),
                    unit("hallas", "hallas", "emos", "/tmp/emos"),
                    # unit("itanium", "itanium", "emos", "/tmp/emos"),),
                unit("localhost", "localhost", "ns", "/tmp/emos/logs"),
                unit("ibis", "ibis", "ns", "/tmp/emos/logs"),
                ))
        else:
            rdir = "/tmp"
            out += (unit("localhost", "localhost", "ns", "/tmp/%s/logs" % user),
                     unit("lxab", "lxab", "serial", udir + "/logs"),
                     unit("lxop", "lxop", "emos", udir + "/logs"),
                     unit("ecgb", "ecgb", "ns", udir + "/logs"),
                     unit("lxc", "lxc", "ns", udir + "/logs"),         )
        return out

def maker():
    return Task("maker").add(
        Variables(ECF_FILES= wdir,
                  ECF_INCLUDE= "/home/ma/emos/def/o/include",
                  WSHOST= "ibis",
                  USER= "emos",
                  SCHOST= "ibis",
                  ECF_OUT= "/vol/lxop_emos_nc/output",),   )

def usage():
    print('''cray.py -h -u [user] -n [node] -s [suite-name] \
 -e: ecflow''')

if __name__ == "__main__":
    global user, udir, ecflow, host
    suites = { "cray": Cray,
               "test": Cray,
           }
    # suite = "test"; node = "eode"
    # argv = cli_proc.process(suites, selection=suite, host=node)
    user = get_username()
    try: host = sys.argv[-1]
    except: host= "localhost"
    ecflow  = False
    ecflow  = True
    suite   = None
    user = get_username()
    opts, args = getopt.getopt(
          sys.argv[1:], "hp:u:es:n:p:", 
          ["help", "port", "user", "ecflow", "suite", "node", "path"])
    print("#MSG: opts, args", opts, args)

    output = None
    verbose = False
    for o, a in opts:
        if   o in ("-n", "--node"):   host = a
        elif o in ("-h", "--help"):   usage(); sys.exit()
        elif o in ("-s", "--suite"):  suite =  a
        elif o in ("-p", "--path"):   path =  a
        elif o in ("-e", "--ecflow"): ecflow = True
        elif o in ("-u", "--user"):
            user = a
        elif o in ("-p", "--port"):
            port = a
        else: print("#ERR: what?", o, a); assert False, "unhandled option"

    import cli_proc

    if suite is None: 
        cli_proc.ip.SELECTION = user
    elif '/' in suite:
        sel = suite
        if '/' == suite[0]: sel = sel[1:]
        if '/' in sel: sel, family = sel.split('/', 1)
        cli_proc.ip.SELECTION = sel
    else: cli_proc.ip.SELECTION = suite

    if ecflow and host not in ("ode", "map", "pikachu"):
        ic.ecf.ECF_MODE = "ecflow"
    else: 
        import sms2ecf
        sms2ecf.ECF_MODE = "sms"
        ic.ecf.ECF_MODE = "sms"

    print("#MSG: ecflow/sms: ", ic.ecf.ECF_MODE)
    global wdir, rdir, fdef, jdir

    ########### SETTINGS ################
    udir = pwd.getpwnam(user).pw_dir
    wdir = udir + "/course/cray"
    rdir = "/sc1/tcwork/emos/logs"
    fdef = wdir + "/cray.def"
    jdir = "/tmp/%s/cray" % user
    ddef = udir + "/course/cray"
    fdef = ddef + "/cray.def"
    idir = "/home/ma/emos/def/cray/include"
    if user == "map" or host in ["map", ]:
        wdir = "/home/ma/map/course/cray"
    elif user == "emos" or host in ["ode", "od3", "eode"]:
        jdir="/vol/lxop_emos_nc/output"
        rdir="/vol/lxop_emos_nc/output"
        udir="/vol/lxop_emos_nc/output"

    print("##", host, suite)
    argv = cli_proc.process(suites, selection=suite, host=host, proc=False)
    try:    os.makedirs(wdir)
    except: pass # one line on purpose
    try:    os.makedirs(jdir)
    except: pass

    create_task()
    create_smhi()



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

import ecflow
import sys
import os
import ecflow_test_util as Test

if __name__ == "__main__":

    Test.print_test_start(os.path.basename(__file__))
 
    flag = ecflow.Flag()
    assert str(flag) == "", "expected empty string, for an empty flag"
         
    flag_list =  flag.list() # flag_list is of type FlagTypeVec
    my_flag_list = [ ecflow.FlagType.force_abort,
                     ecflow.FlagType.user_edit,
                     ecflow.FlagType.task_aborted,
                     ecflow.FlagType.edit_failed,
                     ecflow.FlagType.jobcmd_failed,
                     ecflow.FlagType.killcmd_failed,
                     ecflow.FlagType.statuscmd_failed,
                     ecflow.FlagType.no_script,
                     ecflow.FlagType.killed,
                     ecflow.FlagType.status,
                     ecflow.FlagType.late,
                     ecflow.FlagType.message,
                     ecflow.FlagType.byrule,
                     ecflow.FlagType.queuelimit,
                     ecflow.FlagType.wait,
                     ecflow.FlagType.locked,
                     ecflow.FlagType.zombie,
                     ecflow.FlagType.no_reque,
                     ecflow.FlagType.archived,
                     ecflow.FlagType.restored,
                     ecflow.FlagType.threshold,
                     ecflow.FlagType.sigterm,
                     ecflow.FlagType.log_error,
                     ecflow.FlagType.checkpt_error,
                     ecflow.FlagType.remote_error
                    ]
    print("Flag list:")
    for flg in flag_list: print("flag ",flag.type_to_string(flg))
    print("My   list:")
    for flg in my_flag_list: print("flag ",flag.type_to_string(flg))

    assert len(flag_list) == len(my_flag_list), "expected flag list have changed"
    expected_flags = "force_aborted,user_edit,task_aborted,edit_failed,ecfcmd_failed,killcmd_failed,statuscmd_failed" \
                     ",no_script,killed,status,late,message,by_rule,queue_limit,task_waiting,locked,zombie,no_reque" \
                     ",archived,restored,threshold,sigterm,log_error,checkpt_error,remote_error";

    #Set *ALL* the flags
    for flg in flag_list:
       flag.set( flg )
       assert flag.is_set( flg ),"expected flag %r to be set" % flag.type_to_string(flg)
    
    # clear *ALL* the flags
    assert str(flag) == expected_flags, "Expected flags \n{0}\nbut found \n{1}".format(str(flag),expected_flags)
    for flg in flag_list:
        flag.clear(flg)
        assert not flag.is_set( flg ),"expected flag {0} not to be set".format(flag.type_to_string(flg))

    # test reset
    for flg in flag_list:
       flag.set( flg )
    flag.reset()
    for flg in flag_list:
        assert not flag.is_set( flg ),"expected flag %r not to be set" % flag.type_to_string(flg)
        
    #===========================================================================
    # Setting flag type
    #===========================================================================
    for flg in my_flag_list:
        flag.set( flg )
        print("current flag:",flag)
        assert flag.is_set( flg ),"expected flag %r to be set" % flag.type_to_string(flg)

    #===========================================================================
    # Node:  get the flag, should be empty for a newly created node.
    #===========================================================================
    suite = ecflow.Suite("s1")
    flag = suite.get_flag()
    for flg in flag_list:
        assert not flag.is_set( flg ),"expected flag %r not to be set" % flag.type_to_string(flg)
    assert str(flag) == "", "expected empty string, for an empty flag"
    
    #assert False
    print("All tests pass")



#####1####/2####/3####/4####/5####/6####/7####/8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http:#www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#####1####/2####/3####/4####/5####/6####/7####/8

#  Test that simulation works

import os,fnmatch
import ecflow
import ecflow_test_util as Test

def simulate_defs_with_time():
    #suite test_time_series
    #  family family
    #      task t1
    #         time 00:00 23:00 04:00  # should run 6 times 00:00 04:00 08:00 12:00 16:00 20:00
    #  endfamily
    #endsuite
    print("simulate_defs_with_time()")
    defs = ecflow.Defs()
    suite = defs.add_suite("test_time_series")
    clock = ecflow.Clock(1, 1, 2011, False)       # day,month, year, hybrid make test start at midnight, otherwise current time used
    suite.add_clock(clock)
    family = suite.add_family("family")
    task = family.add_task("t1")
    ts = ecflow.TimeSeries(ecflow.TimeSlot(0,0), ecflow.TimeSlot(23,0), ecflow.TimeSlot(4,0), True)
    task.add_time( ecflow.Time(ts) )
    task.add_verify( ecflow.Verify(ecflow.State.complete, 6) ) # expect task to complete 6 times
         
    theResult = defs.simulate()
    assert len(theResult) == 0,  "Expected simulation to return without any errors, but found:\n" + theResult
    
    print("  simple check for state change time")
    print("   iso_extended:",task.get_state_change_time())
    print("   iso_extended:",task.get_state_change_time("is0_extended"))
    print("   iso         :",task.get_state_change_time("iso"))
    print("   simple      :",task.get_state_change_time("simple"))
    print("   rubbish     :",task.get_state_change_time("rubbish"))
 
    os.remove("test_time_series.def.log")
    
def simulate_deadlock():
    # This simulation is expected to fail, since we have a deadlock/ race condition
    
    print ("simulate_deadlock")
    # suite dead_lock
    #  family family
    #    task t1
    #          trigger t2 == complete
    #    task t2
    #          trigger t1 == complete
    #   endfamily
    # endsuite
    
    defs = ecflow.Defs()
    suite = defs.add_suite("dead_lock")
    fam = suite.add_family("family")
    fam.add_task("t1").add_trigger("t2 == complete")
    fam.add_task("t2").add_trigger("t1 == complete")
    
    theResult = defs.simulate()
    assert len(theResult) != 0, "Expected simulation to return errors, but found none" 
    print(theResult)
           
    os.remove("dead_lock.def.log")
    os.remove("defs.depth")
    os.remove("defs.flat")

def test_time_series():
    print ("Simulator:: ...test_time_series")

    # suite suite
    #  clock real <sunday>
    #    family family
    #       task t1
    #       time 00:30 23:59 04:00  # should run 6 times 00:30 4:30 8:30 12:30 16:30 20:30
    #      endfamily
    #endsuite

    # Initialise real clock on a Monday, such that task should _only_ run
    # on Monday since we are using a hybrid clock
    clock = ecflow.Clock(12, 10, 2009, False)     # day,month, year, hybrid 12 October 2009 was a Monday

    theDefs = ecflow.Defs()
    suite = theDefs.add_suite("test_time_series")
    suite.add_clock( clock )

    fam = suite.add_family( "family" )
    task = fam.add_task("t")
    
    task.add_time("00:30 23:59 04:00");
    task.add_verify( ecflow.Verify(ecflow.State.complete, 6) ) # expect task to complete 6 times

    theResult = theDefs.simulate()
    assert len(theResult) == 0,  "Expected simulation to return without any errors, but found:\n" + theResult

    os.remove("test_time_series.def.log")
    

if __name__ == "__main__":
    Test.print_test_start(os.path.basename(__file__))
 
    simulate_defs_with_time()
    simulate_deadlock()
    test_time_series()
    
    # traverse the CSim test data, make sure python simulation matches c++
    workspace_dir = ecflow.File.source_dir()
    csim_test_data = workspace_dir + "/CSim/test/data/good_defs"
    print (csim_test_data)
    for path in Test.all_files(csim_test_data,'*.def'):
        print (path)
        theDefs = ecflow.Defs(path)
        theResult = theDefs.simulate()
        assert len(theResult) == 0,  "Expected simulation to return without any errors, but found:\n" + theResult

    # remove generated .log files
    for path in Test.all_files(".",'*.log'):
        os.remove(path)
        
    print("All Tests pass")



#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
# Name        :
# Author      : Avi
# Revision    : $Revision: #10 $
#
# Copyright 2009- ECMWF.
# This software is licensed under the terms of the Apache Licence version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8

import os

# clean up any file after running the online tutorial examples
# Place in separate exception, otherwise not all files will be removed
try: os.remove("test.def")
except: pass

try: os.remove("TestSuite.def")
except: pass

try: os.remove("SuiteBuilder.def")
except: pass

try: os.remove("server.defs")   
except: pass
